"""Divide LaTeX files into parts"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/29_latex.divide.ipynb.

# %% auto 0
__all__ = ['SECOND_PARAMETER_PATTERN', 'SECOND_PARAMETER_PATTERN_WITH_OPTIONAL_STAR', 'THIRD_PARAMETER_PATTERN',
           'THIRD_PARAMETER_PATTERN_WITH_OPTIONAL_STAR', 'SECTION_LIKE_PATTERN', 'ENVIRONMENT_PATTERN',
           'DEFAULT_ENVIRONMENTS_TO_NOT_DIVIDE_ALONG', 'NoDocumentNodeError', 'find_document_node',
           'environment_names_used', 'numbered_newtheorems_counters_in_preamble', 'numberwithins_in_preamble',
           'display_names_of_environments', 'get_node_from_simple_text', 'text_from_node', 'swap_numbers_invoked',
           'divide_latex_text']

# %% ../../nbs/29_latex.divide.ipynb 2
from itertools import product
from os import PathLike
import re
from typing import Optional, Union

from pylatexenc.latexwalker import (
    LatexWalker, LatexEnvironmentNode, LatexMacroNode, LatexNode
)

from pylatexenc.latex2text import (
    LatexNodes2Text
)
import regex

from ..helper.files_and_folders import text_from_file

from ..helper.latex.comments import remove_comments
from .formatting import replace_commands_in_latex_document, replace_input_and_include, custom_commands
from .preamble import divide_preamble, replace_inclusion_of_style_file_with_code


# %% ../../nbs/29_latex.divide.ipynb 4
# matches `\newtheorem{theorem}{Theorem}`, `\newtheorem{proposition}[theorem]{Proposition}`
# does not match `\newtheorem{theorem}{Theorem}[Section]`

# SECOND_PARAMETER_PATTERN = re.compile(
#     # r'\\newtheorem\s*\{\s*(\w+)\s*\}\s*(\[\s*(\w+)\s*\])?\s*\{\s*(.*)\s*\}')
#     r'\\newtheorem\s*\{\s*(\w+)\s*\}\s*(\[\s*(\w+)\s*\])?\s*\{\s*(.*)\s*\}(?!\s*\[\s*(\w+)\s*\])')
SECOND_PARAMETER_PATTERN = regex.compile(
    r'\\newtheorem\s*'
    r'\{((?>[^{}]+|\{(?2)\})*)\}'
    r'\s*(\[\s*(\w+)\s*\])?\s*'
    r'\{((?>[^{}]+|\{(?2)\})*)\}'
    r'(?!\s*\[\s*(\w+)\s*\])',
    regex.MULTILINE)

SECOND_PARAMETER_PATTERN_WITH_OPTIONAL_STAR = regex.compile(
    r'\\newtheorem\*?\s*'
    r'\{((?>[^{}]+|\{(?2)\})*)\}'
    r'\s*(\[\s*(\w+)\s*\])?\s*'
    r'\{((?>[^{}]+|\{(?2)\})*)\}'
    r'(?!\s*\[\s*(\w+)\s*\])',
    regex.MULTILINE)

# matches `\newtheorem{theorem}{Theorem}`, `\newtheorem{theorem}{Theorem}[Section]`,
# does not match `\newtheorem{proposition}[theorem]{Proposition}`
# THIRD_PARAMETER_PATTERN = re.compile(
#     r'\\newtheorem\s*\{\s*(\w+)\s*\}\s*\{\s*(.*)\s*\}\s*(\[\s*(\w+)\s*\])?')
THIRD_PARAMETER_PATTERN = regex.compile(
    r'\\newtheorem\s*'
    r'\{((?>[^{}]+|\{(?2)\})*)\}'
    r'\s*'
    r'\{((?>[^{}]+|\{(?2)\})*)\}'
    r'\s*'
    r'(\[\s*(\w+)\s*\])?',
    regex.MULTILINE)

THIRD_PARAMETER_PATTERN_WITH_OPTIONAL_STAR = regex.compile(
    r'\\newtheorem\*?\s*'
    r'\{((?>[^{}]+|\{(?2)\})*)\}'
    r'\s*'
    r'\{((?>[^{}]+|\{(?2)\})*)\}'
    r'\s*'
    r'(\[\s*(\w+)\s*\])?',
    regex.MULTILINE)

# matches \section{title}, \subsection{title}, \subsubsection{title}, \section*{title}, etc.
SECTION_LIKE_PATTERN = regex.compile(
        r'\\(?:section|subsection|subsubsection)\s*(?:\[.*\])?(\*)?\s*'
        r'\{((?>[^{}]+|\{(?2)\})*)\}',
        regex.MULTILINE)
    

# matches \begin{theorem},
ENVIRONMENT_PATTERN = regex.compile(
        r'\\begin\s*'
        r'\{((?>[^{}]+|\{(?1)\})*)\}',
        regex.MULTILINE)


# %% ../../nbs/29_latex.divide.ipynb 9
class NoDocumentNodeError(Exception):
    """Exception raised when a LatexEnvironmentNode corresponding to the document 
    environment is expected in a LaTeX string, but no such node exists.
    
    **Attributes**
    - text - str
        - The text in which the document environment is not found.
    """
    
    def __init__(self, text):
        self.text = text
        super().__init__(
            f"The following text does not contain a document environment:\n{text}")



# %% ../../nbs/29_latex.divide.ipynb 10
def find_document_node(
        text: str, # LaTeX str
        document_environment_name: str = "document" # The name of the document environment.
        ) -> LatexEnvironmentNode:
    """Find the `LatexNode` object for the main document in `text`.
    
    **Raises**
    - NoDocumentNodeError
        - If document environment node is not detected.
    """
    w = LatexWalker(text)
    nodelist, _, _ = w.get_latex_nodes(pos=0)
    for node in nodelist:
        if node.isNodeType(LatexEnvironmentNode)\
                and node.environmentname == document_environment_name:
            return node
    raise NoDocumentNodeError(text)

# %% ../../nbs/29_latex.divide.ipynb 21
def environment_names_used(
        text: str # LaTeX document
        ) -> set[str]: # The set of all environment names used in the main document.
    """Return the set of all environment names used in the main document
    of the latex code.
    """
    document_node = find_document_node(text)
    return {node.environmentname for node in document_node.nodelist
            if node.isNodeType(LatexEnvironmentNode)}        

# %% ../../nbs/29_latex.divide.ipynb 30
def _search_counters_by_pattern(
        preamble: str,
        newtheorem_regex: re.Pattern, # This is supposed to be a regex that detects and captures parameters of `\newtheorem` commands.
        counter_group: int # This depends on which `newtheorem_regex` is used, and is either 3 or 4. 
        ) -> dict[str, str]: # The 
    """
    Capture the newly defined theorem-like environment names as well as the
    counters that they belong to
    
    This is a helper function for `numbered_newtheorems_counters_in_preamble`.
    
    """
    counters = {}
    for match in newtheorem_regex.finditer(preamble):
        env_name = match.group(1)
        counter = match.group(counter_group)
        # If no counter was specified, use the environment name as the counter
        if counter is None:
            counter = env_name
        counters[env_name] = counter
    return counters

# %% ../../nbs/29_latex.divide.ipynb 33
def _article_is_amsart_or_article(
        preamble: str # The preamble with no comments
        ):
    """
    helper function of `numbered_newtheorems_counters_in_preamble`.
    """
    return bool(re.search(r'\\documentclass\s*(\[\s*(.*?)\s*\])?\s*\{\s*(amsart|article)\}', preamble))


# %% ../../nbs/29_latex.divide.ipynb 35
def numbered_newtheorems_counters_in_preamble(
        document: str, # The LaTeX document
        add_equation_counter: Optional[bool] = None, # Determines whether or not the `equation` environment will have a counter added when a `newthoerem` command for the `equation` environment is not explicitly invoked in the preamble. If `None`, then the counter is added if the article is of class `amsart` or `article`. If `True`, then the counter is added. If `False`, then the counter is not added.
        ) -> dict[str, tuple[str, Union[str, None]]]: # The keys are the command names of the environments. The value a key is a tuple `(<counter>, <reset_by_counter>)`, where `<counter>` is the counter that the environment belongs to, which can be custom defined or predefined in LaTeX, and `<reset_by_counter>` is a counter whose incrementation resets the # counter of the environment, if available. 
    r"""Return the dict specifying the numbered `\newtheorem` command invocations

    Assumes that

    - invocations of the `\newtheorem` command are exclusively in the
    preamble of the LaTeX document.
    - theorem-like environments are defined using the `\newtheorem` command.
    - no environments of the same name are defined twice.
    - There is at most one invocation of `\theoremstyle` or `\newtheorem` in each line.

    This function does not take into account `\numberwithins` being used.
    The `numberwithins_in_preamble` function accounts for invocations of
    the `\numberwithins` command instead.

    The `equation` environment (and other related environments, such as `eqnarray`)
    seems to be included in documents of
    the class `amsart` or `article` (i.e. documents which invoke
    `\documentclass{amsart}` or `\documentclass{article}`,
    possibly with some optional arguments).
    The `equation` environment (and other related environments) is accordingly included
    in the output
    of this function if the document is of the class `amsart` and 
    `add_equation_counter` is not specified, set to `None`.

    This function uses two separate regex patterns, one to detect the invocations
    of `\newtheorem` in which the optional parameter is the second parameter and
    one to detect those in which the optional parameter is the third parameter.


    """
    preamble, _ = divide_preamble(document)
    preamble = remove_comments(preamble)
    # TODO: maybe use the `regex` package instead of `re` with a recursive
    # balanced-curly braces detecting regex.
    commands_and_counters = _combine_second_and_third_paramter_results(preamble)
    if 'equation' not in commands_and_counters and (
            add_equation_counter == True or
            add_equation_counter is None and _article_is_amsart_or_article(preamble)):
        commands_and_counters['equation'] = ('equation', None)  
        if 'eqnarray' not in commands_and_counters:
            commands_and_counters['eqnarray'] = ('equation', None)

    return commands_and_counters


def _combine_second_and_third_paramter_results(preamble):
    """
    Inspect invocations of the `\newtheorem` command in the preamble,
    separately dealing with invocations with a third optional parameter vs.
    a second optional parameter.

    helper function of `numbered_newtheorems_counters_in_preamble`.
    """
    # matches `\newtheorem{theorem}{Theorem}`, `\newtheorem{proposition}[theorem]{Proposition}`
    # does not match `\newtheorem{theorem}{Theorem}[Section]`
    # second_parameter_pattern = re.compile(
    #     # In this case, the optional parameter (if any) should not follow the newtheorem.
    #     r'\\newtheorem\s*\{\s*(\w+)\s*\}\s*(\[\s*(\w+)\s*\])?\s*\{\s*(.*)\s*\}(?!\s*\[\s*(\w+)\s*\])')

    # second_parameter_pattern = regex.compile(
    #     r'\\newtheorem\s*'
    #     r'\{((?>[^{}]+|\{(?2)\})*)\}'
    #     r'\s*(\[\s*(\w+)\s*\])?\s*'
    #     r'\{((?>[^{}]+|\{(?2)\})*)\}'
    #     r'(?!\s*\[\s*(\w+)\s*\])',
    #     regex.MULTILINE)

    # matches `\newtheorem{theorem}{Theorem}`, `\newtheorem{theorem}{Theorem}[Section]`,
    # does not match `\newtheorem{proposition}[theorem]{Proposition}`
    # third_parameter_pattern = re.compile(
    #     r'\\newtheorem\s*\{\s*(\w+)\s*\}\s*\{\s*(.*)\s*\}\s*(\[\s*(\w+)\s*\])?')

    # third_parameter_pattern = regex.compile(
    #     r'\\newtheorem\s*'
    #     r'\{((?>[^{}]+|\{(?2)\})*)\}'
    #     r'\s*'
    #     r'\{((?>[^{}]+|\{(?2)\})*)\}'
    #     r'\s*'
    #     r'(\[\s*(\w+)\s*\])?')


    second_results = _search_counters_by_pattern(preamble, SECOND_PARAMETER_PATTERN, 3)
    third_results = _search_counters_by_pattern(preamble, THIRD_PARAMETER_PATTERN, 4)
    to_return = {}
    for environment_name, counter in second_results.items():
        to_return[environment_name] = (counter, None)
    for environment_name, reset_counter in third_results.items():
        if environment_name in to_return:
            continue
        to_return[environment_name] = (environment_name, reset_counter)
    return to_return        

# %% ../../nbs/29_latex.divide.ipynb 49
def numberwithins_in_preamble(
        document: str # The LaTeX document
    ) -> dict[str, str]: # The keys are the first arguments of `numberwithin` invocations and the values ar ethe second arguments of `numberwithin` invocations.
    r"""Return the `dict` describing `\numberwithin` commands invoked
    in the preamble of `document`.
    
    Assumes that `\numberwithin` commands are invoked exclusively in the
    preamble.

    See also the `numbered_newtheorems_counter_in_preamble` function,
    which parses invocations of the `\newtheorem` command.
    """
    preamble, _ = divide_preamble(document)
    preamble = remove_comments(preamble)
    pattern = regex.compile(r'\\numberwithin\s*\{\s*(\w+)\s*\}\s*\{\s*(.*)\s*\}')
    numberwithins = {}

    for match in pattern.finditer(preamble):
        environment_to_number = match.group(1)
        environment_to_count = match.group(2)
        numberwithins[environment_to_number] = environment_to_count

    return numberwithins

# %% ../../nbs/29_latex.divide.ipynb 55
def display_names_of_environments(
        document: str # The LaTeX document
        ) -> dict[str, str]:  
    r"""Return the dict specifying the display names for each theorem-like
    environment.

    This function uses two separate regex patterns, one to detect the invocations
    of `\newtheorem`
    in which the optional parameter is the second parameter and one to detect
    those in which the optional parameter is the third parameter.

    Assumes that
    - invocations of the `\newtheorem` command are exclusively in the
    preamble of the LaTeX document.
    - theorem-like environments are defined using the `\newtheorem` command.
    - no environments of the same name are defined twice.

    """
    preamble, _ = divide_preamble(document)
    second_results = _search_display_names_by_pattern(preamble, SECOND_PARAMETER_PATTERN_WITH_OPTIONAL_STAR, 4)
    third_results = _search_display_names_by_pattern(preamble, THIRD_PARAMETER_PATTERN_WITH_OPTIONAL_STAR, 2)
    return second_results | third_results
    

def _search_display_names_by_pattern(
        preamble: str,
        newtheorem_regex: re.Pattern,
        display_name_group: int # This depends on which `newtheorem_regex` is used, and is either 3 or 4. 
        ) -> dict[str, str]:
    """
    Capture the newly defined theorem-like environment names as well as the
    counters that they belong to"""
    display_names = {}
    for match in newtheorem_regex.finditer(preamble):
        env_name = match.group(1)
        display_name = match.group(display_name_group)
        display_names[env_name] = display_name
    return display_names

# %% ../../nbs/29_latex.divide.ipynb 64
def _setup_counters(
        numbertheorem_counters: dict[str, tuple[str, Union[str, None]]], # An output of `numbered_newtheorems_counters_in_preamble`
        ) -> dict[str, int]:
    r"""
    Return a dict whose keys are of counters in the LaTeX document and whose
    values are all `0`. These key-value pairs are used to keep track of
    the numberings of `parts`.

    One special key is the key of the empty string `''`, which counters the
    parts which do not get a numbering, i.e. for most text that lie outside
    of (numbered) environments

    """

    # cf. https://www.overleaf.com/learn/latex/Counters#Default_counters_in_LaTeX
    predefined_counters = [
        'part', # Incremented each time the `\part` command is used. It is not reset automatically and casn only be reset by the user
        'chapter', # Incremeneted each time the `\chapter` command is used.
        'section', # Incremented whenever a new `\section` command is encountered
        'subsection', # Incremented whenever a new `\subsection` command is encountered, reset whenever a new `\section` command is encountered
        'subsubsection', # Incremented whenever a new `\subsubsection` command is encounted, reset whenever a new `\subsection` or `\section` command is encountered
        'paragraph', # Incremeneted whenever a new paragraph is started. Reset whenever a new `\subsubsection`, `\subsection`, or `\section` command is encounted
        'subparagraph', # Incremented each time the `\subparagraph` command is used and reset at the beginning of a new
        'page', # Incremented each time a new page is started in the document
        'equation', # Incremeneted whenever the `\begin{equation}` environment is used. 
        'figure', # Incremented whenever a new `figure` environment is encountered
        'table', # Incremeneted whenever a new `taable` environment is encountered`
        'footnote', 
        'mpfootnote',
        'enumi',
        'enumii',
        'enumiii',
        'enumiv']

    counters = {counter: 0 for _, (counter, reset_counter) in numbertheorem_counters.items()}
    for counter in predefined_counters:
        counters[counter] = 0

    counters[''] = 0
    return counters

# %% ../../nbs/29_latex.divide.ipynb 66
def _setup_numberwithins(
        explicit_numberwithins: dict[str, str],
        numbertheorem_counters: dict[str, tuple[str, Union[str, None]]], # An output of `numbered_newtheorems_counters_in_preamble`.
        ) -> dict[str, str]: # The keys are counters and the values are all counters that the key is immediately numbered within.
    """
    Extracts information of counters that are reset when other counters are
    incremented.

    This is a helper function of `_setup_all_numberwithins` as well as
    `divide_latex_text`.
    """
    builtin_numberwithins = {
        'subsection': 'section',
        'subsubsection': 'subsection',
        'paragraph': 'subsubsection',
        'subparagraph': 'paragraph',
        'enumii': 'enumi',
        'enumiii': 'enumii',
        'enumiv': 'enumiii',
        'part': 'chapter',
        'appendix': 'chapter'
    }
    numberwithins = explicit_numberwithins | builtin_numberwithins

    for environmentname, (counter, reset_by_counter) in numbertheorem_counters.items():
        if reset_by_counter is None:
            continue
        numberwithins[environmentname] = reset_by_counter
    return numberwithins

    

def _setup_all_numberwithins(
        explicit_numberwithins: dict[str, str],
        numbertheorem_counters: dict[str, tuple[str, Union[str, None]]], # An output of `numbered_newtheorems_counters_in_preamble`.
        ) -> dict[str, list[str]]: # The keys are counters and the values are all counters that the key is numbered within.
    """
    This is a helper function of `divide_latex_text`.
    """
    numberwithins = _setup_numberwithins(explicit_numberwithins, numbertheorem_counters)
    all_counters = set()
    for key, value in numberwithins.items():
        all_counters.add(key)
        all_counters.add(value)
    all_numbered_withins = {counter: [] for counter in all_counters}
    for counter_1, counter_2 in product(all_counters, all_counters):
        if _is_numberedwithin(counter_1, counter_2, numberwithins):
            all_numbered_withins[counter_1].append(counter_2)
    return all_numbered_withins


def _is_numberedwithin(
        counter_1, counter_2, numberwithins: dict[str, str]
        ) -> bool:
    """Return `True` if `counter_1` is numbered within `counter_2""" 
    if counter_1 not in numberwithins:
        return False
    elif numberwithins[counter_1] == counter_2:
        return True
    return _is_numberedwithin(
        numberwithins[counter_1], counter_2, numberwithins)


# %% ../../nbs/29_latex.divide.ipynb 68
def _unnumbered_environments(
        numbertheorem_counters: dict[str, tuple[str, Union[str, None]]], # An output of `numbered_newtheorems_counters_in_preamble`
        display_names: dict[str, str]) -> set[str]:
    r"""Return the set of unnumbered theorem-like environments defined by
    `\newtheorem`.

    This is a helper function of `divide_latex_text`.
    """
    return {environment for environment in display_names
            if environment not in numbertheorem_counters}

    

# %% ../../nbs/29_latex.divide.ipynb 70
def _section_title(
        text: str
        ) -> tuple[bool, str]: # The bool is `True` if the section/subsection is numbered (i.e. is `section` or `subsection` as opposed to `section*` or `subsection*`). The `str` is the title of the section or subsection
    """Return the title of a section or subsection from a latex str
    and whether or not the section/subsection is numbered"""

    # Note that the `section` command has the optional argument `toc-title` which appears
    # in the table of contents, cf.
    # http://latexref.xyz/_005csection.html
    # pattern = regex.compile(
    #     r'\\(?:section|subsection|subsubsection)\s*(?:\[.*\])?(\*)?\s*'
    #     r'\{((?>[^{}]+|\{(?2)\})*)\}',
    #     regex.MULTILINE
    # )
    regex_search = regex.search(SECTION_LIKE_PATTERN, text)
    is_numbered = regex_search.group(1) is None
    title = regex_search.group(2)
    return is_numbered, title


# %% ../../nbs/29_latex.divide.ipynb 72
def _is_section_node(node: LatexNode):
    return (node.isNodeType(LatexMacroNode)
            and node.macroname == 'section')

def _is_subsection_node(node: LatexNode):
    return (node.isNodeType(LatexMacroNode)
            and node.macroname == 'subsection')

def _is_subsubsection_node(node: LatexNode):
    return (node.isNodeType(LatexMacroNode)
            and node.macroname == 'subsubsection')

def _is_environment_node(node: LatexNode):
    return node.isNodeType(LatexEnvironmentNode)

def _text_is_of_section_like_node(text: str):
    """Return `True` if `text` represents the text for a section node.

    In principal, this function should act like 
    `_is_section_node or _is_subsection_node or _is_subsubsection_node`
    except that it takes a `str` as its argument instead of a `LatexNode`.
    This function is
    implemented using a regex pattern instead of using
    `_is_section_node` to save time.
    """
    return bool(regex.match(SECTION_LIKE_PATTERN, text.lstrip()))


def _text_is_of_environment_node(text: str):
    """Return `True` if `text` represents an environment node
    (at least at the start).
    
    In principal, this function should act like `_is_environment_node`
    except that it takes a `str` as its argument instead of a `LatexNode`.
    This function is implemented using a regex pattern instead of using
    '_is_environment_node` to save time.
    """
    return bool(regex.match(ENVIRONMENT_PATTERN, text.lstrip()))


def _environment_name_of_text(text: str):
    """Return `True` if `text` represents an environment node
    (at least at the start).
    
    Assumes that `_text_is_of_environment_node(text)` is `True`.
    """
    match = regex.match(ENVIRONMENT_PATTERN, text.lstrip())
    return match.group(1)

# %% ../../nbs/29_latex.divide.ipynb 75
def _is_numbered(
        node: LatexNode,
        numbertheorem_counters: dict[str, str]
        ) -> bool:
    if _is_section_node(node) or _is_subsection_node(node) or _is_subsubsection_node(node):
        is_numbered, _ = _section_title(node.latex_verbatim())
        return is_numbered
    elif _is_environment_node(node):
        return node.environmentname in numbertheorem_counters
    else:
        return False

# %% ../../nbs/29_latex.divide.ipynb 77
def get_node_from_simple_text(
        text: str) -> LatexNode:
    """Return the (first) `LatexNode` object from a str."""
    w = LatexWalker(text)
    nodelist, _, _ = w.get_latex_nodes(pos=0)
    return nodelist[0]


def text_from_node(
        node: LatexNode) -> str:
    """Return the str representing `node`."""
    return node.latex_verbatim()
    # l2t = LatexNodes2Text()
    # if node.isNodeType('text'):
    #     return node.chars
    # else:
    #     full_text = ''
    #     for child_node in node.children:
    #         full_text += text_from_node(child_node)
    #     return full_text
    # return LatexNodes2Text().node_to_text(node)


# %% ../../nbs/29_latex.divide.ipynb 80
def _change_counters(
        node,
        counters,
        numbertheorem_counters: dict[str, str],
        all_numberwithins: dict[str, list[str]]
        ):
    """Preliminarily update the counters for `node`, but not for
    any of its subnodes. This is mostly for
    theoremlike environments and for sectionlike environments.
    
    Helper function to `_process_node`.
    """
    # identify which counter to change
    # TODO
    # Take into consideration unnumbered non-environment node
    # Take into consideration unnumbered environment node
    if _is_environment_node(node):
        if node.environmentname in numbertheorem_counters:
           counter = numbertheorem_counters[node.environmentname][0]
        else:
            counter = None
    elif _is_section_node(node):
        counter = 'section'
    elif _is_subsection_node(node):
        counter = 'subsection'
    elif _is_subsubsection_node(node):
        counter = 'subsubsection'
    else:
        counter = None

    # Section counters seem to only reset subsection counters
    # When the section is numbered, etc., cf. `numbering_example_4...`
    # and `numbering_example_5...` in `nbs\_tests\latex_examples`
    is_numbered = _is_numbered(node, numbertheorem_counters)
    # e.g. `\numberwithin{equation}{section}`` means that `equation` is
    # numbered within `section`, i.e. `equation` is reset whenever
    # `section` is incremeneted

    # if counter is None and not _is_environment_node(node):
    #     counters[''] += 1 
    #     return

    if is_numbered:
        counters[counter] += 1
    for numbered_counter, within_counter in all_numberwithins.items():
        if counter is not None and counter in within_counter:
            counters[numbered_counter] = 0


def _change_counters_antecedently(
        node,
        counters,
        numbertheorem_counters: dict[str, str],
        all_numberwithins: dict[str, list[str]],
        ):
    """Update the counters to account for any environments contained within
    `node`, but not `node` itself.

    This is mostly for theorem-like environments which share a counter with 
    something like the `equation` environment; sometimes a theorem-like
    environment can have an `equation` environment within it. Note that
    the `_process_node` function alreay invokes `_change_counters` on
    `node`, so the counter is already updated for `node` by the time
    `_change_counters_antecedently` is invoked.

    Helper function to `_process_node`.
    """
    queue = [node]
    while queue:
        subnode = queue.pop()
        _update_counter_for_subsubnodes(
            subnode, counters, numbertheorem_counters, all_numberwithins)
        queue.extend(_subsubnodes(subnode))


def _update_counter_for_subsubnodes(
        subnode,
        counters,
        numbertheorem_counters: dict[str, str],
        all_numberwithins: dict[str, list[str]],
        ) -> None: 
    """Iterate through the immediate subnodes of `subnode` to see if the
    counter needs to be updated for any
    
    Helper function to `_change_counters_antecedently`.
    """ 
    if not hasattr(subnode, 'nodelist'):
        return
    for subsubnode in subnode.nodelist:
       _change_counters(subsubnode, counters, numbertheorem_counters, all_numberwithins)
            

def _subsubnodes(
        subnode,
        ) -> list[LatexNode]: 
    """Find subnodes of `subnode` to add to the queue

    Helper function to `_change_counters_antecedently`.
    """ 
    if not hasattr(subnode, 'nodelist'):
        return []
    return subnode.nodelist




# %% ../../nbs/29_latex.divide.ipynb 83
def _node_numbering(
        node: LatexNode,
        numbertheorem_counters: dict[str, str],
        numberwithins: dict[str, str],
        counters: dict[str, int]
        ) -> str: # Just the numbering of the node, no "section/subsection" or displayname
    if _is_section_node(node):
        counter = 'section'
    elif _is_subsection_node(node):
        counter = 'subsection'
    elif _is_environment_node(node):
        counter = numbertheorem_counters[node.environmentname][0]
    return _numbering_helper('', counter, numberwithins, counters)


def _numbering_helper(
        trailing_numbering: str,
        counter: str,
        numberwithins: dict[str, str],
        counters: dict[str, int]
        ) -> str:
    """Recurisve helper function to `_node_numbering`."""
    if counter not in numberwithins and counter not in counters:
        return trailing_numbering
    if counter not in numberwithins and counter in counters and trailing_numbering:
        return f'{counters[counter]}.{trailing_numbering}'
    if counter not in numberwithins and counter in counters and not trailing_numbering:
        return f'{counters[counter]}'

    parent_counter = numberwithins[counter]
    current_count = counters[counter]
    if not trailing_numbering:
        to_pass_to_trailing_numbering = str(current_count)
    else:
        to_pass_to_trailing_numbering = f'{current_count}.{trailing_numbering}'

    return _numbering_helper(
        to_pass_to_trailing_numbering,
        parent_counter,
        numberwithins,
        counters)
    

# %% ../../nbs/29_latex.divide.ipynb 85
def _title(
        node: LatexNode,
        numbertheorem_counters: dict[str, str],
        numberwithins: dict[str, str], # An output of _setup_numberwithins
        all_numberwithins: dict[str, list[str]], # An output of all_numberwithins
        display_names: dict[str, str],
        counters: dict[str, int],
        swap_numbers: bool):
    """Return the title of a node based on the count in
    `counters`"""
    numbered = _is_numbered(node, numbertheorem_counters)

    if (_is_section_node(node) or _is_subsection_node(node)
            or _is_subsubsection_node(node)):
        return _title_for_section_subsection_subsubsection_node(
            node, counters, bool)
    # if _is_section_node(node) and numbered:
    #     _, title = _section_title(node.latex_verbatim())
    #     return f"{counters['section']}. {title}"
    # if _is_section_node(node) and not numbered:
    #     _, title = _section_title(node.latex_verbatim())
    #     return title 

    # if _is_subsection_node(node) and numbered:
    #     _, title = _section_title(node.latex_verbatim())
    #     return f"{counters['section']}.{counters['subsection']}. {title}"
    # if _is_subsection_node(node) and not numbered:
    #     _, title = _section_title(node.latex_verbatim())
    #     return title

    if _is_environment_node(node):
        return _title_for_environment_node(
            node, numbertheorem_counters, numberwithins,
            display_names, counters, swap_numbers)


def _title_for_section_subsection_subsubsection_node(
        node: LatexNode,
        counters: dict[str, int],
        numbered: bool
    ):
    _, title = _section_title(node.latex_verbatim())
    if not numbered:
        return title
    if _is_section_node(node):
        return f"{counters['section']}. {title}"
    elif _is_subsection_node(node):
        return f"{counters['section']}.{counters['subsection']}. {title}"
    else: # _is_subsubsection_node(node):
        return f"{counters['section']}.{counters['subsection']}.{counters['subsubsection']}. {title}"

    # if _is_section_node(node) and numbered:
    #     _, title = _section_title(node.latex_verbatim())
    #     return f"{counters['section']}. {title}"
    # if _is_section_node(node) and not numbered:
    #     _, title = _section_title(node.latex_verbatim())
    #     return title 

    # if _is_subsection_node(node) and numbered:
    #     _, title = _section_title(node.latex_verbatim())
    #     return f"{counters['section']}.{counters['subsection']}. {title}"
    # if _is_subsection_node(node) and not numbered:
    #     _, title = _section_title(node.latex_verbatim())
    #     return title


def _title_for_environment_node(
        node: LatexNode,
        numbertheorem_counters: dict[str, str],
        numberwithins: dict[str, list[str]],
        display_names: dict[str, str],
        counters: dict[str, int],
        swap_numbers: bool):
    """Return the title of an environment node.
    If the node is not that of an theorem-like environment, then 
    
    """
    numbered = _is_numbered(node, numbertheorem_counters)
    # TODO: see what happens when environments are numbered within
    # sections vs. subsections
    if not numbered:
        numbering = None
    else:
        numbering = _node_numbering(
            node, numbertheorem_counters, numberwithins, counters)
    
    environment = node.environmentname
    if environment in display_names:
        display_name = display_names[environment]
    else:
        display_name = environment
    if not numbered:
        return display_name
    elif swap_numbers:
        return f'{numbering}. {display_name}.'
    else:
        return f'{display_name} {numbering}.'
        

# %% ../../nbs/29_latex.divide.ipynb 87
def swap_numbers_invoked(
        preamble: str
        ) -> bool: # 
    """Returns `True` if `\swapnumbers` is in the preamble.

    Assume that a mention of `\swapnumbers` is an actual invocation.
    """
    preamble = remove_comments(preamble)
    return '\swapnumbers' in preamble

# %% ../../nbs/29_latex.divide.ipynb 89
def _node_warrants_own_part(
        node, environments_to_not_divide_along: list[str],
        accumulation: str, parts: list[tuple[str, str]]) -> bool:
    """Return `True` if `node` warrants making a new part to be added in `parts`.

    This is a helper function for `_process_node`. When this function returns
    `True`, the `accumulation` should be considered for appending to `parts`
    and the node should also be appended to `parts
    """
    if _is_section_node(node) or _is_subsection_node(node) or _is_subsubsection_node(node):
        return True
    elif not _is_environment_node(node):
        return False
    # Is environment node from here and below.
    if len(parts) == 0 and accumulation.strip() == '':
        return True
    return node.environmentname not in environments_to_not_divide_along

# %% ../../nbs/29_latex.divide.ipynb 91
def _node_is_proof_immediately_following_a_theorem_like_environment(
        node, accumulation, parts, display_names) -> bool:
    """Return `True` if `node` is that of a proof environment that immediately
    follows a theorem-like environment.
    
    This is a helper function for `_process_node`.
    """
    if not _is_environment_node(node):
        return False
    if not node.environmentname == 'proof':
        return False
    if not len(parts) > 0:
        return False
    if accumulation.strip() != '':
        return False
    if not _text_is_of_environment_node(parts[-1][1]):
        return False
    return _environment_name_of_text(parts[-1][1]) in display_names
    # previous_node = get_node_from_simple_text(parts[-1][1])
    # if not _is_environment_node(previous_node):
    #     return False
    # return previous_node.environmentname in display_names


def _node_is_nonspecial_following_a_sectionlike_node(
        node, environments_to_not_divide_along, accumulation, parts) -> bool:
    """Return `True` if `node` is that of a non-environment and non-setionlike
    node that immediately follows a sectionlike (section, subsection,
    or subsubsection) node.
    
    This is a helper function for `_process_node`.
    """
    if ((_is_environment_node(node) and node.environmentname not in environments_to_not_divide_along)
            or _is_section_node(node)
            or _is_subsection_node(node)
            or _is_subsubsection_node(node)):
        return False
    if len(parts) == 0:
        return False
    if accumulation.strip() != '':
        return False
    # previous_node = get_node_from_simple_text(parts[-1][1])
    # return _is_section_node(previous_node) or _is_subsection_node(previous_node) or _is_subsubsection_node(previous_node)
    return _text_is_of_section_like_node(parts[-1][1])
    

# %% ../../nbs/29_latex.divide.ipynb 93
DEFAULT_ENVIRONMENTS_TO_NOT_DIVIDE_ALONG = [
    'align', 'align*', 'diagram', 'displaymath', 'displaymath*', 'enumerate', 'eqnarray', 'eqnarray*',
    'equation', 'equation*', 'gather', 'gather*', 'itemize', 'label',
    'multiline', 'multiline*', 'multline', 'multline*',
    'proof', 'quote', 'tabular', 'table', ]
def divide_latex_text(
        document: str, 
        # environments_to_divide_along: list[str], # A list of the names of environments that warrant a new note
        # numbered_environments: list[str], # A list of the names of environments which are numbered in the latex code. 
        dir: Optional[PathLike], # The directory where the included files and style files are to be found.
        environments_to_not_divide_along: list[str] = DEFAULT_ENVIRONMENTS_TO_NOT_DIVIDE_ALONG, # A list of the names of the environemts along which to not make a new note, unless the environment starts a section (or the entire document).
        replace_commands_in_document_first: bool = True,  # If `True`, invoke `replace_commands_in_latex_document` on `document` to first replace custom commands (in the document minus the preamble) before starting to divide the document.
        repeat_replacing_commands: int = -1,  # If `replace_commands_in_document_first` is `True`, then this is passed as the `repeat` argument into the invocation of `replace_commands_in_latex_document`.
        ) -> list[tuple[str, str]]: # Each tuple is of the form `(note_title, text)`, where `note_title` often encapsulates the note type (i.e. section/subsection/display text of a theorem-like environment) along with the numbering and `text` is the text of the part. Sometimes `title` is just a number, which means that `text` is not of a `\section` or `\subsection` command and not of a theorem-like environment.
    r"""Divide LaTeX text to convert into Obsidian.md notes.

    Assumes that the counters in the LaTeX document are either the
    predefined ones or specified by the `\newtheorem` command.

    Proof environments are assigned to the same parts their prcededing
    theorem-like environments, if available.

    TODO: Implement counters specified by `\newcounter`, cf. 
    https://www.overleaf.com/learn/latex/Counters#LaTeX_commands_for_working_with_counters.
    """
    numbertheorem_counters = numbered_newtheorems_counters_in_preamble(document)
    explicit_numberwithins = numberwithins_in_preamble(document)
    numberwithins = _setup_numberwithins(explicit_numberwithins, numbertheorem_counters)
    all_numberwithins = _setup_all_numberwithins(explicit_numberwithins, numbertheorem_counters)
    display_names = display_names_of_environments(document)
    counters = _setup_counters(numbertheorem_counters)
    unnumbered_environments = _unnumbered_environments(
        numbertheorem_counters, display_names)
    # Eventually gets returned
    preamble, main_document = divide_preamble(document)
    preamble = replace_inclusion_of_style_file_with_code(preamble, dir)
    preamble_commands = custom_commands(preamble)
    if replace_commands_in_document_first:
        main_document = replace_input_and_include(
            main_document, dir, preamble_commands,
            repeat_replacing_commands)
        # main_document = replace_commands_in_latex_document(document, repeat_replacing_commands)
    document_node = find_document_node(main_document)
    swap_numbers = swap_numbers_invoked(preamble)
    parts = []
    # "Accumulates" a "part" until text that should comprise a new part is encountered
    accumulation = '' 
    for node in document_node.nodelist:
        accumulation = _process_node(
            node, environments_to_not_divide_along, accumulation,
            numbertheorem_counters,
            numberwithins, all_numberwithins, counters,
            display_names, swap_numbers, parts)
    _append_non_environment_accumulation_to_parts_if_non_empty(
        accumulation, counters, parts)
    return parts


def _process_node(
        node, environments_to_not_divide_along, accumulation,
        numbertheorem_counters,
        numberwithins, all_numberwithins, counters,
        display_names, swap_numbers, parts) -> str:
    """
    Update `accumulation`, `counter`, and `parts` based on the contents of `node`.

    Also return 'accumulation` to update it.

    This is a helper function for `divide_latex_text`.

    """
    # If node is a proof immediately following a theorem-like environment
    # Then add it to said theorem-like environment
    _change_counters(
        node, counters, numbertheorem_counters, numberwithins)
    if (_node_is_proof_immediately_following_a_theorem_like_environment(
            node, accumulation, parts, display_names)
        or _node_is_nonspecial_following_a_sectionlike_node(
            node, environments_to_not_divide_along, accumulation, parts)):
        parts[-1][1] += node.latex_verbatim()
    elif _node_warrants_own_part(
            node, environments_to_not_divide_along, accumulation, parts):
        accumulation =  _append_non_environment_accumulation_to_parts_if_non_empty(
            accumulation, counters, parts)
        
        title = _title(
            node, numbertheorem_counters, numberwithins, all_numberwithins,
            display_names, counters, swap_numbers).strip()
        title = title.replace('\n', '') 
        parts.append([title, node.latex_verbatim()])
    else:
        accumulation += node.latex_verbatim()
        # In _change_counters`, the '' counter is incremented by default.
        # This offsets the incorrectly incrementation.
    _change_counters_antecedently(node, counters, numbertheorem_counters, all_numberwithins)
    return accumulation


def _append_non_environment_accumulation_to_parts_if_non_empty(
        accumulation: str, counters, parts):
    """Append accumulation to `parts` if `accumulation` is nonempty
    and return the updated `accumulation` """
    if accumulation.strip() != '':
        counters[''] += 1
        parts.append([str(counters['']).strip(), accumulation.strip()])
        return ''
    else:
        return accumulation.strip()








"""Detect whether packages are installed and install them during run-time as needed."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/star.ipynb.

# %% auto 0
__all__ = ['check_package_installed', 'check_package_version_with_version_attribute', 'install_package',
           'ensure_package_installed']

# %% ../../nbs/star.ipynb 2
import importlib.util
import subprocess
import sys

# %% ../../nbs/star.ipynb 4
def check_package_installed(
        package_name: str
        ) -> bool:
    if importlib.util.find_spec(package_name) is not None:
        return True
        # print(f"{package_name} is installed.")
    else:
        # print(f"{package_name} is not installed.")
        return False


# %% ../../nbs/star.ipynb 6
def check_package_version_with_version_attribute(package_name):
    """Checks and returns the version of the specified package if available.
    
    Args:
        package_name (str): The name of the package to check.

    Returns:
        str or None: The version of the package or None if the package is not available.
    """
    # Check if the package is already imported
    if package_name in sys.modules:
        pkg = sys.modules[package_name]
    else:
        try:
            # Import the package dynamically
            pkg = __import__(package_name)
        except ImportError:
            return None  # Return None if the package is not available

    # Attempt to return the version attribute
    return getattr(pkg, '__version__', None)

# %% ../../nbs/star.ipynb 10
def install_package(package_name: str) -> None:
    """Install a package using pip."""
    subprocess.check_call([sys.executable, '-m', 'pip', 'install', package_name])

def ensure_package_installed(package_name: str) -> None:
    """Check if a package is installed and install it if not."""
    if check_package_installed(package_name):
        print(f"{package_name} is already installed.")
    else:
        print(f"{package_name} is not installed. Installing...")
        install_package(package_name)
        print(f"{package_name} has been installed.")




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/46_helper.files_and_folders.ipynb.

# %% auto 0
__all__ = ['existing_path', 'file_existence_test', 'path_name_no_ext', 'path_no_ext', 'text_from_file', 'files_of_format_sorted',
           'file_is_compressed', 'uncompress_file', 'get_download_path', 'get_huggingface_cache_dir']

# %% ../../nbs/46_helper.files_and_folders.ipynb 1
import bz2
import errno
import gzip
import lzma
import os
from os import PathLike
from pathlib import Path
import platform
import sys
import tarfile
from typing import Optional
import zipfile

from deprecated import deprecated
import glob
from natsort import natsorted


# %% ../../nbs/46_helper.files_and_folders.ipynb 2
if sys.platform.startswith('win'):
    import winreg
else:
    winreg = None

# %% ../../nbs/46_helper.files_and_folders.ipynb 6
def existing_path(
        path: PathLike,  # A file or directory path. Either absolute or relative to `relative_to`.
        relative_to: Optional[PathLike] = None  # Path to the directory that `file` is relative to.  If `None`, then `path` is an absolute path.
        ) -> Path: # The path formed by `relative_to` adjoined with `path`.  Defaults to `None`
    """Returns a path relative to a specified path as an absolute path
    that exists.

    **Raises**

    - `FileNotFoundError`
        - If `relative_to` is not `None` but does not exist, or if
        `file` does not exist.
    - `ValueError`
        - If `relative_to` is not `None` and yet not an absolute path, or
        if `relative_to` is `None` at yet `path` is not an absolute path.
    
    **Notes**
    - This function may add the string `'\\\\?\\'` in front, which identifies
    very long paths.
    """
    if relative_to is not None:
        if not os.path.isabs(relative_to):
            raise ValueError(
                f'The parameter `relative_to` is expected to be an'
                f' absolute path, but it is not: {relative_to}')
        if not os.path.exists(relative_to):
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), relative_to)
        path = Path(relative_to) / path
    elif not os.path.isabs(path):
        raise ValueError(
            f'The parameter `path` is expected to be an absolute path,'
            f' but it is not: {path}')
    if not os.path.exists(path) and platform.system() == 'Windows':
        path = f'\\\\?\\{str(path)}'  # For long file names
    if not os.path.exists(path):
        raise FileNotFoundError(
            errno.ENOENT, os.strerror(errno.ENOENT), path)
    return Path(path)


@deprecated(reason="The function has been renamed to `existing_path`")
def file_existence_test(
        path: PathLike,  # A file or directory path. Either absolute or relative to `relative_to`.
        relative_to: Optional[PathLike] = None  # Path to the directory that `file` is relative to.  If `None`, then `path` is an absolute path.
        ) -> Path: # The path formed by `relative_to` adjoined with `path`.  Defaults to `None`
    """
    **Deprecated. Use `existing_path` instead.**
    
    Returns a path relative to a specified path as an absolute path
    that exists.

    **Raises**
    - `FileNotFoundError`
        - If `relative_to` is not `None` but does not exist, or if
        `file` does not exist.
    
    **Notes**
    - This function may add the string `'\\\\?\\'` in front, which identifies
    very long paths.
    """
    if relative_to is not None:
        if not os.path.isabs(relative_to):
            raise ValueError(
                f'The parameter `relative_to` is expected to be an'
                f' absolute path, but it is not: {relative_to}')
        if not os.path.exists(relative_to):
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), relative_to)
        path = Path(relative_to) / path
    elif not os.path.isabs(path):
        raise ValueError(
            f'The parmaeter `path` is expected to be an absolute path,'
            f' but it is not: {path}')
    if not os.path.exists(path) and platform.system() == 'Windows':
        path = f'\\\\?\\{str(path)}'  # For long file names
    if not os.path.exists(path):
        raise FileNotFoundError(
            errno.ENOENT, os.strerror(errno.ENOENT), path)
    return Path(path)

# %% ../../nbs/46_helper.files_and_folders.ipynb 19
def path_name_no_ext(
        path: PathLike # The path of the file or directory. This may be absolute or relative to any directory.
        ) -> str: # The name of the file or directory without the extension.
    """Return the name of a file or directory from its path without the
    extension.
    
    The file or directory does not have to exist.
    """
    name_with_extension = os.path.basename(path)
    return os.path.splitext(name_with_extension)[0]

# %% ../../nbs/46_helper.files_and_folders.ipynb 26
def path_no_ext(
    path: PathLike # The path of the file or directory. This may be absolute or relative to any directory.
    ) -> str: # The path of the file or directory without the extension. If `path` is a path to a directory, then the output should be essentially the same as `path`.
    """Returns the path of a file or directory without the extension.
    
    The file or directory does not have to exist.
    """
    return os.path.splitext(str(path))[0]

# %% ../../nbs/46_helper.files_and_folders.ipynb 30
def text_from_file(
        path: PathLike, # The absolute path of the file.
        encoding: str = 'utf8' # The encoding of the file to be read. Defaults to `'utf8'`.
        ) -> str: # The entire text from a file
    """Return the entire text from a file.

    Assuems that the file can be encoded in the specified `encoding`
    """
    with open(path, 'r', encoding=encoding) as file:
        text = file.read()
        file.close()
    return text

# %% ../../nbs/46_helper.files_and_folders.ipynb 34
def files_of_format_sorted(
        directory: PathLike, # The directory in which to find the files
        extension: str = 'txt' # Extension of the files to find. Defaults to 'txt'.
        ) -> list[str]:
    """Return a list of path str of files in the directory (but not subdirectories)
    sorted via `natsort`.
    """
    return natsorted(glob.glob(str(Path(directory) / f'*.{extension}')))

# %% ../../nbs/46_helper.files_and_folders.ipynb 38
def file_is_compressed(
        filename: str
        ):
    # Define a set of common compressed file extensions
    compressed_extensions = {
        '.zip',
        '.gz',
        '.tar',
        '.tar.gz',
        '.tgz',
        '.bz2',
        '.xz',
        '.7z',
        '.rar',
        '.z',
    }

    # Get the file extension
    _, file_extension = os.path.splitext(filename)

    # Check if the file extension is in the set of compressed extensions
    return file_extension.lower() in compressed_extensions

# %% ../../nbs/46_helper.files_and_folders.ipynb 40
def uncompress_file(
        file_path: PathLike,
        verbose: bool = False
        ):
    # Check the file extension
    file_path = str(file_path)
    _, file_extension = os.path.splitext(file_path)
    uncompressed_files = []

    try:
        if file_extension == '.zip':
            with zipfile.ZipFile(file_path, 'r') as zip_ref:
                zip_ref.extractall(os.path.dirname(file_path))
                uncompressed_files = [os.path.join(os.path.dirname(file_path), name) for name in zip_ref.namelist()]
            if verbose:
                print(f"Uncompressed {file_path} into {os.path.dirname(file_path)}")

        elif file_extension in ['.tar', '.tar.gz', '.tgz', '.tar.bz2', '.tbz']:
            with tarfile.open(file_path, 'r:*') as tar_ref:
                tar_ref.extractall(os.path.dirname(file_path))
                uncompressed_files = [os.path.join(os.path.dirname(file_path), name) for name in tar_ref.getnames()]
            if verbose:
                print(f"Uncompressed {file_path} into {os.path.dirname(file_path)}")

        elif file_extension == '.gz':
            output_file_path = file_path[:-3]  # Remove the .gz extension
            with gzip.open(file_path, 'rb') as gz_file:
                with open(output_file_path, 'wb') as out_file:
                    out_file.write(gz_file.read())
            uncompressed_files.append(output_file_path)
            if verbose:
                print(f"Uncompressed {file_path} into {output_file_path}")

        elif file_extension == '.bz2':
            output_file_path = file_path[:-4]  # Remove the .bz2 extension
            with bz2.open(file_path, 'rb') as bz2_file:
                with open(output_file_path, 'wb') as out_file:
                    out_file.write(bz2_file.read())
            uncompressed_files.append(output_file_path)
            if verbose:
                print(f"Uncompressed {file_path} into {output_file_path}")

        elif file_extension == '.xz':
            output_file_path = file_path[:-3]  # Remove the .xz extension
            with lzma.open(file_path, 'rb') as xz_file:
                with open(output_file_path, 'wb') as out_file:
                    out_file.write(xz_file.read())
            uncompressed_files.append(output_file_path)
            if verbose:
                print(f"Uncompressed {file_path} into {output_file_path}")

        else:
            if verbose:
                print(f"Unsupported file format: {file_extension}")

    except Exception as e:
        if verbose:
            print(f"An error occurred while uncompressing {file_path}: {e}")

    return uncompressed_files

# import os
# import zipfile
# import tarfile
# import gzip
# import bz2
# import lzma
# from pathlib import Path

# def uncompress_file(
#         file_path: PathLike,
#         verbose: bool = False
#         ):
#     file_path = Path(file_path)
#     output_dir = file_path.parent
#     uncompressed_files = []

#     try:
#         if file_path.suffix == '.zip':
#             with zipfile.ZipFile(file_path, 'r') as zip_ref:
#                 zip_ref.extractall(output_dir)
#                 uncompressed_files = [output_dir / name for name in zip_ref.namelist()]

#         elif file_path.suffix in ['.tar', '.gz', '.tgz', '.bz2', '.tbz']:
#             if file_path.suffix == '.gz' and not tarfile.is_tarfile(file_path):
#                 # Handle single gzipped file
#                 with gzip.open(file_path, 'rb') as gz_file:
#                     content = gz_file.read()
#                     # Try to get the original filename from the gzip header
#                     original_name = gz_file.name
#                     if original_name:
#                         output_file = output_dir / Path(original_name).name
#                     else:
#                         output_file = output_dir / file_path.stem
#                     output_file.write_bytes(content)
#                 uncompressed_files.append(output_file)
#             else:
#                 # Handle tar archives (including .tar.gz, .tgz, .tar.bz2, .tbz)
#                 with tarfile.open(file_path, 'r:*') as tar_ref:
#                     tar_ref.extractall(output_dir)
#                     uncompressed_files = [output_dir / name for name in tar_ref.getnames()]

#         elif file_path.suffix == '.bz2':
#             with bz2.open(file_path, 'rb') as bz2_file:
#                 content = bz2_file.read()
#                 output_file = output_dir / file_path.stem
#                 output_file.write_bytes(content)
#             uncompressed_files.append(output_file)

#         elif file_path.suffix == '.xz':
#             with lzma.open(file_path, 'rb') as xz_file:
#                 content = xz_file.read()
#                 output_file = output_dir / file_path.stem
#                 output_file.write_bytes(content)
#             uncompressed_files.append(output_file)

#         else:
#             if verbose:
#                 print(f"Unsupported file format: {file_path.suffix}")

#         if verbose:
#             print(f"Uncompressed {file_path} into {output_dir}")

#     except Exception as e:
#         if verbose:
#             print(f"An error occurred while uncompressing {file_path}: {e}")

#     return uncompressed_files


# %% ../../nbs/46_helper.files_and_folders.ipynb 45
def get_download_path() -> str:
    """
    Return the user's download folder
    """
    if os.name == 'nt':  # For Windows
        sub_key = r'SOFTWARE\Microsoft\Windows\CurrentVersion\Explorer\Shell Folders'
        downloads_guid = '{374DE290-123F-4565-9164-39C4925E467B}'
        with winreg.OpenKey(winreg.HKEY_CURRENT_USER, sub_key) as key:
            location = winreg.QueryValueEx(key, downloads_guid)[0]
        return location
    else:  # For Unix-based systems (Linux, macOS)
        return os.path.join(os.path.expanduser('~'), 'Downloads')

# %% ../../nbs/46_helper.files_and_folders.ipynb 47
def get_huggingface_cache_dir():
    # Determine the cache directory
    cache_dir = os.environ.get("HF_HOME") or os.environ.get("XDG_CACHE_HOME")
    if not cache_dir:
        home = Path.home()
        cache_dir = home / ".cache" / "huggingface" / "hub"
    else:
        cache_dir = Path(cache_dir) / "huggingface" / "hub"
    return cache_dir



"""Helper functions"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/00_helper.ipynb.

# %% ../../nbs/00_helper.ipynb 2
from __future__ import annotations



# %% auto 0
__all__ = ['substring_generator', 'sublist_generator']

# %% ../../nbs/00_helper.ipynb 4
def substring_generator(input_string: str):
    length = len(input_string)
    for i in range(length):
        for j in range(i + 1, length + 1):
            yield input_string[i:j]


def sublist_generator(input_list: list):
    length = len(input_list)
    for i in range(length):
        for j in range(i + 1, length + 1):
            yield input_list[i:j]



"""Functions for gathering and processing data to train and for using  ML models to link notation notes with one another."""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../../nbs/34_markdown.obsidian.machine_learning.notation_linking.ipynb.

# %% auto 0
__all__ = ['data_from_notation_notes', 'data_points_for_reference', 'text_from_data_point', 'prediction_by_model',
           'prediction_by_model_via_datapoint', 'auto_add_link_to_notation_note',
           'auto_add_link_to_notation_note_via_data_point']

# %% ../../../../../nbs/34_markdown.obsidian.machine_learning.notation_linking.ipynb 5
from os import PathLike
import random
import re
from typing import Optional, Union

from ....markdown.file import MarkdownFile, MarkdownLineEnum
from ...links import MARKDOWNLINK_CAPTURE_PATTERN
from ..information_notes import reference_of_information_note
from ..notation.in_standard_information_note import notation_notes_linked_in_see_also_section
from ..notation.parse import parse_notation_note
from ..note_processing import process_standard_information_note
from ..note_type import note_is_of_type, type_of_note, PersonalNoteTypeEnum
from ..notes import notes_linked_in_notes_linked_in_note
from .notation_summarization import _notation_note_has_auto_summary_tag
from ...vault import VaultNote

# %% ../../../../../nbs/34_markdown.obsidian.machine_learning.notation_linking.ipynb 9
# TODO: test
def _init_args_for_data_from_notation_notes(
        origin_notation_note: VaultNote, # The notation note which potentially uses the notation introduced by `relied_notation_note`. In particular, there potentially ought to be a link to `relied_notation_note` in `origin_notation_note`.`
        relied_notation_note: VaultNote, # See the description for `origin_notation_note`.`,
        origin_parsed: Optional[tuple] = None, # The output of `parse_notation_note` applied to `origin_notation_note`
        relied_parsed: Optional[tuple] = None, # The output of `parse_notation_note` applied to `relied_notation_note`
        reference_name: Optional[str] = None, # The name of the reference folder in the vault from which the two notation notes comes from. If `None`, this is computed "on-the-fly" based on the reference of the main note of main_of_origin, see `reference_of_information_note`
        main_of_origin_content: Optional[str] = None, # The content of `main_of_origin`, i.e. the output of `process_standard_information_note(MarkdownFile.from_vault_note(main_of_origin))`. If None, this is computed "one-the-fly".
        main_of_relied_content: Optional[str] = None, # The content of `main_of_relied`, i.e. the output of `process_standard_information_note(MarkdownFile.from_vault_note(main_of_relied))`. If None, this is computed "one-the-fly".
        information_notes_of_reference: Optional[list[VaultNote]] = None, # The standard information notes for the reference folder in order (as arranged in the index notes of the reference folder)
    ) -> tuple[str, str, Union[list[VaultNote], None], VaultNote, VaultNote, str, str, list[VaultNote]]  :

    if origin_parsed is None:
        origin_parsed = parse_notation_note(origin_notation_note)
    if relied_parsed is None:
        relied_parsed = parse_notation_note(relied_notation_note)

    main_of_origin, main_of_relied = origin_parsed[2], relied_parsed[2]
    if reference_name is None and main_of_origin is not None:
        reference_name = reference_of_information_note(main_of_origin)

    if main_of_origin_content is None and main_of_origin is not None:
        main_of_origin_content = process_standard_information_note(
            MarkdownFile.from_vault_note(main_of_origin))
    if main_of_relied_content is None and main_of_relied is not None:
        main_of_relied_content = process_standard_information_note(
            MarkdownFile.from_vault_note(main_of_relied))

    if information_notes_of_reference is None:
        # TODO Implement the initialization of `information_notes_of_reference`.
        information_notes_of_reference = information_notes_of_reference
    return (
        origin_parsed, relied_parsed, reference_name, main_of_origin_content,
        main_of_relied_content, information_notes_of_reference)



# %% ../../../../../nbs/34_markdown.obsidian.machine_learning.notation_linking.ipynb 10
# TODO: test
def _notat_str(
        meta: Union[dict, None],
        notat_str: str
    ) -> str:
    r"""Get the closest thing to the original latex string from which a notation note
    is derived from based on the note's YAML frotmatter metadata.

    If the metadata has a `latex_in_original` field, then return its first entry.
    Otherwise, return `notat_str`, which is the string such that the notation note
    starts with the formatting `$<notat_str>$ [[<link_to_main_note>|denotes]]...`.

    This is a helper function to `data_from_notation_notes`.
    """
    if meta and 'latex_in_original' in meta and len(meta['latex_in_original']) > 0:
        return meta['latex_in_original'][0]
    else:
        return notat_str


def _origin_links_to_relied(
        linked_notat_notes: list[tuple], # One of the outputs of `parse_notation_note`
        origin_content: MarkdownFile,  # One of the outputs of `parse_notation_note`
        relied_notation_note: VaultNote
        ) -> bool:
    r"""Returns `True` if the origin notation note links to `relied_notation_note`.

    Helper function to `data_from_notation_notes`.
    
    This function checks both the trailing list of links as well as any links within
    the content of the origin notation note.
    """

    linked_note_names = []
    for _, notat_note_name in linked_notat_notes:
        if notat_note_name.endswith('.md'):
            notat_note_name = notat_note_name[:-3]
        linked_note_names.append(notat_note_name)
        # if notat_note_name == relied_notation_note.name:
        #     return True
    linked_note_names.extend(
        _linked_note_names_from_content(str(origin_content)))
        
    for note_name in linked_note_names:
        if note_name == relied_notation_note.name:
            return True
    return False


def _adjust_content(
        include_content: bool, 
        content: str,
        meta: Union[dict, None],
        ) -> str:
    r"""
    Helper function to `data_from_notation_notes`.

    Sometimes, the content of a notation note is autogenerated by an ML model
    and hence can be unreliable. This function is used to exclude
    autogenerated content.
    """
    if not include_content or (
            meta is not None and 'tags' in meta and '_auto/notation_summary' in meta['tags']):
        return ""
    return content


def _linked_note_names_from_content(content: str):
    linked_note_names = []
    for match in re.findall(MARKDOWNLINK_CAPTURE_PATTERN, content):
        link_name = match[1]
        if link_name.endswith('.md'):
            link_name = link_name[:-3]
        linked_note_names.append(link_name)
    return linked_note_names
        

def _linked_notat_note_names_from_content(content: str, vault: PathLike):
    linked_note_names = _linked_note_names_from_content(content)
    linked_notation_note_names = []
    for linked_note_name in linked_note_names:
        note = VaultNote(vault, name=linked_note_name, update_cache=False)
        if note_is_of_type(note, PersonalNoteTypeEnum.NOTATION_NOTE):
            linked_notation_note_names.append(note.name)
    return linked_notation_note_names

# %% ../../../../../nbs/34_markdown.obsidian.machine_learning.notation_linking.ipynb 11
# TODO: test
def data_from_notation_notes(
        origin_notation_note: VaultNote, # The notation note which potentially uses the notation introduced by `relied_notation_note`. In particular, there potentially ought to be a link to `relied_notation_note` in `origin_notation_note`.`
        relied_notation_note: VaultNote, # See the description for `origin_notation_note`.`,
        include_origin_content: bool, # If `True`, include the content of `origin_notation_note`, i.e. a summary of what the notation introduced by this note means.
        include_relied_content: bool, # If `True`, include the content of `relied_notation_note`, i.e. a summary of what the notation introduced by this note means.
        origin_parsed: Optional[tuple] = None, # The output of `parse_notation_note` applied to `origin_notation_note`
        relied_parsed: Optional[tuple] = None, # The output of `parse_notation_note` applied to `relied_notation_note`
        reference_name: Optional[str] = None, # The name of the reference folder in the vault from which the two notation notes comes from. If `None`, this is computed "on-the-fly" based on the reference of the main note of main_of_origin, see `reference_of_information_note`
        main_of_origin_content: Optional[str] = None, # The content of `main_of_origin`, i.e. the output of `process_standard_information_note(MarkdownFile.from_vault_note(main_of_origin))`. If None, this is computed "one-the-fly".
        main_of_relied_content: Optional[str] = None, # The content of `main_of_relied`, i.e. the output of `process_standard_information_note(MarkdownFile.from_vault_note(main_of_relied))`. If None, this is computed "one-the-fly".
        information_notes_of_reference: Optional[list[VaultNote]] = None, # The standard information notes for the reference folder in order (as arranged in the index notes of the reference folder)
    ) -> tuple[str, str, str, str, str, str, str, str, str, str, str, str, str, bool]:
    """
    Obtain data for a single pair of notation notes.

    Assumes that

    - `origin_notation_note` and `relied_notation_note` have the same `vault`
      attribute.
    - `origin_parsed` and `relied_parsed` are respectively the outputs of
      `parse_notation_note` applied to `reference_name` if specified.
    - `reference_name` is the correct output of `reference_of_information_note`
      applied to `main_of_origin` and that this output is the same as that when
      applied to `main_of_relied`.
    - `main_of_origin_content` and `main_of_relied_content` are the outputs of
      `process_standard_information_note(MarkdownFile.from_vault_note(main_of_origin))`
      and
      `process_standard_information_note(MarkdownFile.from_vault_note(main_of_relied))`
      respectively if they are specified.
    - `information_notes_of_reference` correctly lists the standard information
      notes from the reference in the vault of name `reference_name`.

    **Returns**
    - tuple[str, str, str, str, str, str, str, str, str, str, str, str, str, bool] -
      Consists of the following:
        1. The name of the reference from which the notation notes and main information
           notes come from,
        2. The name of `origin_notation_note`,
        3. The name of `relied_notation_note`,
        4. The name of the main information note of `origin_notation_note`,
        5. The name of the main information note of `relied_notation_note`,
        6. The (processed) content of `main_of_origin_content`,
        7. The (processed) content of `main_of_relied_content`,
        8. The content of `origin_notation_note`
        9. The content of `relied_notation_note`
        10. The first entry in the `latex_in_original` field in the YAML frontmatter
           metadata or, if unavailable, the notation that is summarized in
           `origin_notation_note`
        11. The first entry in the `latex_in_original` field in the YAML frontmatter
           metadata or, if unavailable, the notation that is summarized in
           `relied_notation_note`
        12. The notation that is summarized in `origin_notation_note`
        13. The notation that is summarized in `relied_notation_note`
        14. `True` if `origin_notation_note` links to `relied_notation_note`.
            `False` otherwises


    """
    (origin_parsed, relied_parsed, reference_name, main_of_origin_content,
     main_of_relied_content, information_notes_of_reference
     ) = _init_args_for_data_from_notation_notes(
         origin_notation_note, relied_notation_note, 
         origin_parsed, relied_parsed,
         reference_name, main_of_origin_content,
         main_of_relied_content, information_notes_of_reference)
     
    (origin_meta, origin_notat_str, main_of_origin_name,
     origin_content, linked_notat_notes
     ) = parse_notation_note(origin_notation_note)
    (relied_meta, relied_notat_str, main_of_relied_name, relied_content, _
     ) = parse_notation_note(relied_notation_note)
    origin_content, relied_content = str(origin_content), str(relied_content)
    origin_content = _adjust_content(include_origin_content, origin_content, origin_meta)
    relied_content = _adjust_content(include_relied_content, relied_content, relied_meta)

    return (
        reference_name, origin_notation_note.name, relied_notation_note.name,
        main_of_origin_name, main_of_relied_name,
        main_of_origin_content, main_of_relied_content,
        origin_content, relied_content,
        _notat_str(origin_meta, origin_notat_str),
        _notat_str(relied_meta, relied_notat_str),
        origin_notat_str, relied_notat_str,
        _origin_links_to_relied(linked_notat_notes, origin_content, relied_notation_note)
        )



# %% ../../../../../nbs/34_markdown.obsidian.machine_learning.notation_linking.ipynb 12
def data_points_for_reference(
        reference_index_note: VaultNote, # The index note for the reference from which to draw the data.
        return_notation_note_parsings: bool = False, # If `True`, return the outputs of `parse_notation_note` applied to the notation notes in the reference folder 
        ) -> Union[list[tuple], tuple[list[tuple], dict[str, tuple]]]:
    r"""Compile data points for notation note linking based on the information
    notes and notation notes in a reference folder.

    "Positive" linking data points are relatively rare in comparison to "Negative"
    data points, so "Negative" data points are randomly sampled (although the random
    samples will redundantly include "Positive" data as well).

    Note that it makes sense to draw data exclusively within each "reference" ---
    notations tend to have dependencies within the same mathematical text.

    **Returns**
    - `Union[list[tuple], tuple[list[tuple], dict[str, tuple]]]`
        - Either 1. a list of tuples --- in this case, each tuple is a
          "data point" and is an output of `data_from_notation_notes` ---
          or 2. the list of tuples along with a dict whose keys are
          the names of the notation notes and whose values are the
          outputs of `parse_notation_note` applied to these notation notes.
    """
    info_notes = notes_linked_in_notes_linked_in_note(reference_index_note, as_dict=False)
    info_notes = [note for note in info_notes if note.exists() and note_is_of_type(note, PersonalNoteTypeEnum.STANDARD_INFORMATION_NOTE)]
    notat_notes = []
    for info_note in info_notes:
        notat_notes.extend(notation_notes_linked_in_see_also_section(info_note, info_note.vault))

    vault = reference_index_note.vault
    info_notes_and_processed_content = {}
    for note in info_notes:
        try:
            info_notes_and_processed_content[note.name] = str(process_standard_information_note(MarkdownFile.from_vault_note(note), vault))
        except TypeError as e:
            print(f"An error occurred while trying to process the following note: {note.name}")
            info_notes_and_processed_content[note.name] = note.text()
    # info_notes_and_processed_content = {
    #     note.name: str(process_standard_information_note(
    #         MarkdownFile.from_vault_note(note), vault))
    #     for note in info_notes}
    notation_notes_and_parsed = {
        notat_note.name: parse_notation_note(notat_note) for notat_note in notat_notes}

    # Notation notes whose summaries are written and not auto-generated
    confirmed_summary_notat_note_names = [
        notat_note for notat_note, parsed in notation_notes_and_parsed.items()
        if not _notation_note_has_auto_summary_tag(parsed[0])]

    # Get all positive pairs of notation notes
    positive_linked_notat_note_pairs = _positive_pairs_of_notation_notes(
        confirmed_summary_notat_note_names, notation_notes_and_parsed, vault)
    data_points = _positive_data_points(
        reference_index_note, positive_linked_notat_note_pairs,
        notation_notes_and_parsed, info_notes_and_processed_content)
    data_points.extend(_sample_data_points(
        reference_index_note, len(data_points)*4,
        notat_notes, confirmed_summary_notat_note_names, notation_notes_and_parsed,
        info_notes_and_processed_content))
    if return_notation_note_parsings:
        return data_points, notation_notes_and_parsed
    else:
        return data_points



def _positive_pairs_of_notation_notes(
        confirmed_summary_notat_note_names: list[str],
        notation_notes_and_parsed: dict[str, tuple],
        vault: PathLike
        ) -> list[tuple[str, str]]:
    r"""Return the pairs `(<origin_notat_note_name>, <linked_notat_note_name>)`
    where `origin_notat_note_name` is the name of a notation note whose notation
    summary is "confirmed" (i.e. written and not autogenerated) and where
    `linked_notat_note_name` is the name of a notation note linked by the
    notation note with name `origin_notat_note_name`.
    """
    positive_linked_notat_note_pairs = []
    for notat_note_name in confirmed_summary_notat_note_names:
        parsed = notation_notes_and_parsed[notat_note_name]
        for _, linked_notat_note_name in parsed[4]:
            if linked_notat_note_name.endswith('.md'):
                linked_notat_note_name = linked_notat_note_name[:-3]
            positive_linked_notat_note_pairs.append(
                (notat_note_name, linked_notat_note_name))
        for linked_notat_note_name in _linked_notat_note_names_from_content(
                str(parsed[3]), vault):
            positive_linked_notat_note_pairs.append(
                (notat_note_name, linked_notat_note_name))

    return positive_linked_notat_note_pairs
    

def _positive_data_points(
        reference_index_note: VaultNote,
        positive_linked_notat_note_pairs: list[tuple[str, str]],
        notation_notes_and_parsed: dict[str, tuple],
        info_notes_and_processed_content: dict[str, str]
        ) -> list[tuple]:
    r"""Gather data points for the pairs `(<origin_notat_note_name>, <linked_notat_note_name>)`
    where `origin_notat_note_name` is the name of a notation note whose notation
    summary is "confirmed" (i.e. written and not autogenerated) and where
    `linked_notat_note_name` is the name of a notation note linked by the
    notation note with name `origin_notat_note_name`.
    """
    vault = reference_index_note.vault
    data_points = []
    for origin_notat_note_name, linked_notat_note_name in positive_linked_notat_note_pairs:
        try:
            origin_notat_note = VaultNote(vault, name=origin_notat_note_name, update_cache=False)
            linked_notat_note = VaultNote(vault, name=linked_notat_note_name, update_cache=False)
            main_of_origin_name = notation_notes_and_parsed[origin_notat_note_name][2]
            main_of_linked_name = notation_notes_and_parsed[linked_notat_note_name][2]
            if (main_of_origin_name is None or main_of_linked_name is None
                    or main_of_origin_name not in info_notes_and_processed_content
                    or main_of_linked_name not in info_notes_and_processed_content):
                continue

            for include_origin_content, include_linked_content in [(True, True), (True, False), (False, True), (False, False)]:
                data_points.append(data_from_notation_notes(
                    origin_notat_note, linked_notat_note,
                    include_origin_content, include_linked_content, 
                    notation_notes_and_parsed[origin_notat_note_name], notation_notes_and_parsed[linked_notat_note_name],
                    reference_index_note.name[7:], 
                    info_notes_and_processed_content[main_of_origin_name], info_notes_and_processed_content[main_of_linked_name],
                    None  # TODO: Pass proper argument for information_notes_of_reference
                    ))
        except Exception as e:
            print('An error has occurred while gathering data:')
            print(e)
    return data_points


def _sample_data_points(
        reference_index_note: VaultNote,
        num_samples: int, 
        # info_notes: list[VaultNote],
        notat_notes: list[VaultNote],
        confirmed_summary_notat_note_names: list[VaultNote],
        notation_notes_and_parsed: dict[str, tuple],
        info_notes_and_processed_content: dict[str, str]
        ) -> list[tuple]:
    r"""Randomly sample pairs `(<origin_notat_note_name>, <relied_notat_note_name>)`
    """
    # num_samples = min(num_samples, len(confirmed_summary_notat_note_names))
    origins = random.choices(confirmed_summary_notat_note_names, k=num_samples)
    origins = [VaultNote(reference_index_note.vault, name=name, update_cache=False) for name in origins]
    relieds = random.choices(notat_notes, k=num_samples)
    data_points = []
    for origin, relied in zip(origins, relieds):
        main_of_origin_name = notation_notes_and_parsed[origin.name][2]
        main_of_relied_name = notation_notes_and_parsed[relied.name][2]
        if (main_of_origin_name is None or main_of_relied_name is None
                or main_of_origin_name not in info_notes_and_processed_content
                or main_of_relied_name not in info_notes_and_processed_content):
            continue
        for include_origin_content, include_relied_content in [(True, True), (True, False), (False, True), (False, False)]:
            data_points.append(data_from_notation_notes(
                origin, relied,
                include_origin_content, include_relied_content, 
                notation_notes_and_parsed[origin.name], notation_notes_and_parsed[relied.name],
                reference_index_note.name[7:], 
                info_notes_and_processed_content[main_of_origin_name], info_notes_and_processed_content[main_of_relied_name],
                None  # TODO: Pass proper argument for information_notes_of_reference
                ))
    return data_points

# %% ../../../../../nbs/34_markdown.obsidian.machine_learning.notation_linking.ipynb 13
def text_from_data_point(
        data_point: tuple  # An output of `data_from_notation_notes`.
        ) -> str:
    r"""
    Format a data point to present it as a str.
    """
    text = (f"latex_in_original for origin_notation_note: {data_point[9]}\n"
            f"latex_in_original for relied_notation_note: {data_point[10]}\n"
            f"Notation for origin_notation_note: {data_point[11]}\n"
            f"Notation for relied_notation_note: {data_point[12]}\n"
            f"Summary for origin_notation_note: {data_point[7]}\n"
            f"Summary for relied_notation_note: {data_point[8]}\n\n\n\n")
    content_origin = f"Content for main note of origin_notation_note: {data_point[5]}"
    content_relied = _content_relied(data_point[5], data_point[6])
    return f"{text}{content_origin}\n\n\n\n{content_relied}"


def _content_relied(
        main_of_origin_content: str,
        main_of_relied_content: str,
        ) -> str:
    if main_of_origin_content == main_of_relied_content:
        return f"Content for main note of relied_notation_note: same as that of main note of origin_notation_note" 
    else:
        return f"Content for main note of relied_notation_note: {main_of_relied_content}"

# %% ../../../../../nbs/34_markdown.obsidian.machine_learning.notation_linking.ipynb 15
def prediction_by_model(
        origin_notation_note: VaultNote, # 
        relied_notation_note: VaultNote, #
        learn, # The fastai textlearner which makes the prediction. 
        include_origin_content: bool, # If `True`, include the content of `origin_notation_note`, i.e. a summary of what the notation introduced by this note means.
        include_relied_content: bool, # If `True`, include the content of `relied_notation_note`, i.e. a summary of what the notation introduced by this note means.
        origin_parsed: Optional[tuple] = None, # The output of `parse_notation_note` applied to `origin_notation_note`
        relied_parsed: Optional[tuple] = None, # The output of `parse_notation_note` applied to `relied_notation_note`
        reference_name: Optional[str] = None, # The name of the reference folder in the vault from which the two notation notes comes from. If `None`, this is computed "on-the-fly" based on the reference of the main note of main_of_origin, see `reference_of_information_note`
        main_of_origin_content: Optional[str] = None, # The content of `main_of_origin`, i.e. the output of `process_standard_information_note(MarkdownFile.from_vault_note(main_of_origin))`. If None, this is computed "one-the-fly".
        main_of_relied_content: Optional[str] = None, # The content of `main_of_relied`, i.e. the output of `process_standard_information_note(MarkdownFile.from_vault_note(main_of_relied))`. If None, this is computed "one-the-fly".
        information_notes_of_reference: Optional[list[VaultNote]] = None, # The standard information notes for the reference folder in order (as arranged in the index notes of the reference folder)
        ) -> bool: # `True` if `model` predicts that `origin_notat_note` depends on the notation summarized by `relied_notat_note` and hence should link to `relied_notat_note`.
    r"""
    Predict whether a notation note depends on the notation
    summarized by another notation note.

    See `data_from_notation_notes` for details on the parameters, except `learn`.

    See also `prediction_by_model_via_datapoint` for an alternative
    function for predictions.
    """
    data_point = data_from_notation_notes(
        origin_notation_note, relied_notation_note, include_origin_content,
        include_relied_content, origin_parsed, relied_parsed, reference_name,
        main_of_origin_content, main_of_relied_content,
        information_notes_of_reference)
    input = text_from_data_point(data_point)
    pred, loss, _ = learn.predict(input)
    return pred == 'True'


def prediction_by_model_via_datapoint(
        learn, # The fastai textlearner which makes the prediction. 
        data_point: tuple, # An output of `data_from_notation_notes`
        ) -> bool:
    r"""

    """
    input = text_from_data_point(data_point)
    pred, loss, _ = learn.predict(input)
    return pred == 'True'

# %% ../../../../../nbs/34_markdown.obsidian.machine_learning.notation_linking.ipynb 16
# TODO: mark the note with an _auto tag and make it so that data collection doesn't pick
# up auto-generated links.
def auto_add_link_to_notation_note(
        origin_notation_note: VaultNote,
        relied_notation_note: VaultNote,
        learn, # The fastai textlearner which makes the prediction. 
        include_origin_content: bool, # If `True`, include the content of `origin_notation_note`, i.e. a summary of what the notation introduced by this note means.
        include_relied_content: bool, # If `True`, include the content of `relied_notation_note`, i.e. a summary of what the notation introduced by this note means.
        origin_parsed: Optional[tuple] = None, # The output of `parse_notation_note` applied to `origin_notation_note`
        relied_parsed: Optional[tuple] = None, # The output of `parse_notation_note` applied to `relied_notation_note`
        reference_name: Optional[str] = None, # The name of the reference folder in the vault from which the two notation notes comes from. If `None`, this is computed "on-the-fly" based on the reference of the main note of main_of_origin, see `reference_of_information_note`
        main_of_origin_content: Optional[str] = None, # The content of `main_of_origin`, i.e. the output of `process_standard_information_note(MarkdownFile.from_vault_note(main_of_origin))`. If None, this is computed "one-the-fly".
        main_of_relied_content: Optional[str] = None, # The content of `main_of_relied`, i.e. the output of `process_standard_information_note(MarkdownFile.from_vault_note(main_of_relied))`. If None, this is computed "one-the-fly".
        information_notes_of_reference: Optional[list[VaultNote]] = None, # The standard information notes for the reference folder in order (as arranged in the index notes of the reference folder)
        ) -> None:
    r"""
    Predict whether `origin_notation_note` depends on the notation
    summarized by `relied_notation_note` and add a link in
    `origin_notation_note` to `relied_notation_note` if so.

    It is recommended to pass precomputed arguments for
    the optional parameters. In particular, passing a precomputed
    argument to `origin_parsed` prevents the link to
    `relied_notation_note` from being redundantly added if such
    a link already exists.

    If a link to `relied_notation_note` is added, then an
    `#_auto/notation_notes_linked` tag is added to
    `origin_notation_note` if such a tag is not already
    present.

    """
    if _origin_notation_note_already_has_link_to_relied(
            relied_notation_note, origin_parsed):
        return
    pred = prediction_by_model(
        origin_notation_note, relied_notation_note, learn,
        include_origin_content,
        include_relied_content, origin_parsed, relied_parsed, reference_name,
        main_of_origin_content, main_of_relied_content,
        information_notes_of_reference)
    if not pred:
        return
    _add_notation_link(origin_notation_note, relied_notation_note, relied_parsed)


def _origin_notation_note_already_has_link_to_relied(
        relied_notation_note: VaultNote,
        origin_parsed: Union[tuple, None],
        ) -> bool:
    r"""
    Return `True` if `relied_notation_note is determined to be linked by
    `origin_notation_note` based on the contents of `origin_parsed`, which
    is the output of `parse_notation_note` appleid to `origin_notation_note`.

    This is a helper function to 
    `automatically_add_bulleted_link_to_other_notation_note`.
    """
    for _, linked_note_name in origin_parsed[4]:
        if linked_note_name == relied_notation_note.name:
            return True
    return False


def _add_notation_link(
        origin_notation_note: VaultNote,
        relied_notation_note: VaultNote,
        relied_parsed: tuple,
        ) -> None:
    r"""
    Add a link in `origin_notation_note` to `relied_notation_note`
    and add the tag `_auto/notation_notes_linked` to `origin_notation_note`.
    """
    if not relied_parsed:
        relied_parsed = parse_notation_note(
            relied_notation_note, relied_notation_note.vault)
    mf = MarkdownFile.from_vault_note(origin_notation_note)
    mf.add_tags('_auto/notation_notes_linked',
                enquote_entries_in_metadata_fields=['latex_in_original'])
    bullet = f'- [{relied_parsed[1]}]({relied_notation_note.name}.md)'
    mf.add_line_to_end(
        {'line': bullet,
         'type': MarkdownLineEnum.UNORDERED_LIST})
    mf.write(origin_notation_note)

# %% ../../../../../nbs/34_markdown.obsidian.machine_learning.notation_linking.ipynb 17
def auto_add_link_to_notation_note_via_data_point(
        learn, # The fastai textlearner which makes the prediction. 
        vault: PathLike, 
        data_point: tuple, # An output of `data_from_notation_notes`
        origin_parsed: Optional[tuple] = None, # Optionally, the output of `parse_notation_note` applied to the `origin_notation_note`.
        relied_parsed: Optional[tuple] = None, # Optionally, the output of `parse_notation_note` applied to the `relied_notation_note`.
        ) -> None:
    origin_notation_note = VaultNote(vault, name=data_point[1])
    relied_notation_note = VaultNote(vault, name=data_point[2])
    if origin_parsed is None:
        origin_parsed = parse_notation_note(origin_notation_note, vault)
    if relied_parsed is None:
        relied_parsed = parse_notation_note(relied_notation_note, vault)
    if _origin_notation_note_already_has_link_to_relied(
            relied_notation_note, origin_parsed):
        return
    pred = prediction_by_model_via_datapoint(learn, data_point)
    if not pred:
        return
    _add_notation_link(
        origin_notation_note, relied_notation_note,
        relied_parsed)



"""Helper functors dealing with HTML tags"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/38_helper.html.ipynb.

# %% auto 0
__all__ = ['html_tag_str', 'find_lt_symbols_without_space_in_math_mode', 'add_space_to_lt_symbols_without_space',
           'remove_html_tags_in_text', 'add_HTML_tag_data_to_raw_text']

# %% ../../nbs/38_helper.html.ipynb 2
import bs4
from bs4 import BeautifulSoup
import random
import re
from typing import Optional, Union


from .regex import latex_indices, find_regex_in_text, replace_string_by_indices


# %% ../../nbs/38_helper.html.ipynb 5
def html_tag_str(
        html_tag: bs4.element.Tag
        ) -> str:
    """
    Return the full string of `html_tag`, accounting for 
    special characters that `bs4` changes
    """
    special_chars = {'&lt;': '<', '&gt;': '>', '&amp;': '&'}
    text_to_return = str(html_tag)
    for special_char, replace_with in special_chars.items():
        text_to_return = text_to_return.replace(special_char, replace_with)
    return text_to_return

# %% ../../nbs/38_helper.html.ipynb 10
def find_lt_symbols_without_space_in_math_mode(
        text: str
        ) -> list[int]: # The index of  
    """
    Return the indices in `text` with math mode less than `<` symbols without
    a space that follows.
    """
    latex_inds = latex_indices(text)
    lt_pattern = re.compile(r'<(?! )')
    inds_of_lt_without_spaces_after = []
    for start, end in latex_inds:
        latex_str = text[start:end]
        relative_lt_inds = find_regex_in_text(latex_str, lt_pattern)
        inds_of_lt_without_spaces_after.extend([
            start + relative_lt_ind for relative_lt_ind, _ in relative_lt_inds
        ])
    return inds_of_lt_without_spaces_after 


# %% ../../nbs/38_helper.html.ipynb 14
def add_space_to_lt_symbols_without_space(
        text: str
        ) -> str:
    """Add space after less than `<` symbols if the symbol
    is not followed by a space.
    """
    lt_wo_space_inds = find_lt_symbols_without_space_in_math_mode(text)
    lt_wo_space_ranges = [(ind, ind+1) for ind in lt_wo_space_inds]
    return replace_string_by_indices(
        text, replace_ranges=lt_wo_space_ranges,
        replace_with=['< '] * len(lt_wo_space_inds))
    

# %% ../../nbs/38_helper.html.ipynb 20
def remove_html_tags_in_text(
        text: str, # The text in which to remove the HTML tags.
        replace_with_attributes: Optional[Union[str, list[str]]] = None, # Attribute(s) within the HTML tags which should be used to replace the text of the tags. If `None`, then the texts are not replaced with the attributes. If multiple attributes are specified, then only one attribute is used to replace the text for each HTML tag (independently at random of other replacements). Each attribute's text has an equal chance of being selected for replacement. Repeats are ignored.
        definitely_replace: bool = False, # If `True` and if a given HTML tag has an attribute specified in `replace_with_attributes`, then the text for that tag will definitely be replaced by the text of one of the attributes. Otherwise, the original text and each attribute's text have an equal chance of being selected.
        seed: int = None # Random seed 
        ) -> tuple[str, list[tuple[bs4.element.Tag, int, int]]]: # The text `removed` without HTML tags and a list whose elements consist of the removed HTML tags and the starting and ending indices of the text corresponding to the removed tags within `removed`.
    """Remove the HTML tags in `text`.

    HTML tags are assumed to be not nested.

    """
    random.seed(seed)
    parsed_soup = BeautifulSoup(text, 'html.parser')
    replace_with_attributes = _init_replace_with_attributes(
        replace_with_attributes)

    position = 0
    replaced_contents = []
    for content in parsed_soup.contents:
        position = _process_content(
            parsed_soup, replace_with_attributes, definitely_replace, content,
            position, replaced_contents)
    text_to_return = html_tag_str(parsed_soup)
    return text_to_return, replaced_contents


def _init_replace_with_attributes(
        replace_with_attributes: Optional[Union[str, list[str]]]
        ) -> set[str]:
    if replace_with_attributes is None:
        replace_with_attributes = []
    elif isinstance(replace_with_attributes, str):
        replace_with_attributes = [replace_with_attributes]
    return set(replace_with_attributes)


def _select_replacement_text(
        content: bs4.element.Tag,
        replace_with_attributes: set[str],
        definitely_replace: bool) -> str:
    if not replace_with_attributes:
        return content.string
    selection_pool = []    
    if not definitely_replace:
        selection_pool.append(content.string)
    for attribute, value in content.attrs.items():
        if attribute not in replace_with_attributes:
            continue
        selection_pool.append(value)
    return random.choice(selection_pool)


def _process_content(
        parsed_soup: BeautifulSoup,
        replace_with_attributes: set[str],
        definitely_replace: bool,
        content,
        position: int,
        replaced_contents: list) -> int:
    
    if not isinstance(content, bs4.element.Tag):
        return position + len(content)
    replacement_text = _select_replacement_text(
        content, replace_with_attributes, definitely_replace)
    
    try:
        replaced_content = content.replace_with(
            parsed_soup.new_string(replacement_text))
    except TypeError as e:
        raise e

    replaced_contents.append((
        replaced_content,
        position,
        position + len(replacement_text)))
    return position + len(replacement_text)
    

# %% ../../nbs/38_helper.html.ipynb 37
def add_HTML_tag_data_to_raw_text(
        text: str, # The text onto which to add HTML tags. This is assumed to contain no HTML tags.
        tags_and_locations: list[tuple[bs4.element.Tag, int, int]] # Each tuple consists of the tag object to add as well as the indices within `text` to. The ranges specified by the tuples are assumed to not overlap with one another.
        ) -> str: # The modification of `text` in which the tags are added at the specified locations; the ranges in `text` are replaced..
    """
    Add specified HTML tags to the specified locations/ranges in `text`.

    See the `add_HTML_tag_data_to_text` function for adding HTML
    tag data to text that may or may not already have HTML tags.
    """
    # sort by starting index
    tags_and_locations = sorted(
        tags_and_locations, key=lambda x: x[1])
    replace_ranges = [(start, end) for _, start, end in tags_and_locations]
    replace_with = [html_tag_str(html_tag) for html_tag, _, _ in tags_and_locations]
    return replace_string_by_indices(text, replace_ranges, replace_with)



"""Deal with the comments of a LaTeX document"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/33_latex.comments.ipynb.

# %% auto 0
__all__ = ['remove_comments']

# %% ../../nbs/33_latex.comments.ipynb 2
import re

# %% ../../nbs/33_latex.comments.ipynb 5
def remove_comments(text: str) -> str:
    # Find all occurrences of the comment pattern %[^\n]*
    return re.sub(r"%[^\n]*", "", text)



"""Functions for gathering and processing data on determining where to divide mathematical text"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../../../nbs/53_markdown.obsidian.personal.machine_learning.note_division.ipynb.

# %% auto 0
__all__ = ['combine_info_notes_in_index_notes']

# %% ../../../../../nbs/53_markdown.obsidian.personal.machine_learning.note_division.ipynb 3
from ...vault import VaultNote

# %% ../../../../../nbs/53_markdown.obsidian.personal.machine_learning.note_division.ipynb 5
def combine_info_notes_in_index_notes(
        index_note: VaultNote,
        ) -> tuple[str, list[int]]: 
    """
    
    """
    return



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/40_helper.numbers.ipynb.

# %% auto 0
__all__ = ['is_number']

# %% ../../nbs/40_helper.numbers.ipynb 2
from typing import Union

# %% ../../nbs/40_helper.numbers.ipynb 4
def is_number(
        x: Union[float, int, complex, str]
        ) -> bool:
    """Return `True` if the input `x` represents a number.
    
    This function is different from Python's built-in `is_numeric`
    function, which returns `True` when all characters of a string
    are digits.
    """
    if isinstance(x, (float, int, complex)):
        return True
    #For the case where string is None
    if x is None:
        return False
    if x and x[0] == '-': x = x[1:]
    return x.replace(".", "1", 1).isdigit()



"""Functions for downloading (the source code of) articles from arXiv"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/49_helper.arxiv.ipynb.

# %% auto 0
__all__ = ['arxiv_id', 'arxiv_search', 'extract_metadata', 'ArxivMetadataEncoder', 'extract_last_names', 'folder_name_for_source',
           'create_acronym', 'file_name_for_pdf', 'analyze_arxiv_tarfile', 'read_gz_file', 'get_tex_filename_from_gz',
           'extract_tex_from_gz', 'download_from_results']

# %% ../../nbs/49_helper.arxiv.ipynb 1
import datetime
import gzip
import json
from typing import Callable, Literal, Optional, Union
import os
from os import PathLike
from pathlib import Path
import re
import tarfile

import arxiv
from arxiv import Client, Search, Result
from pathvalidate import sanitize_filename

from .files_and_folders import file_is_compressed, uncompress_file


# %% ../../nbs/49_helper.arxiv.ipynb 5
def arxiv_id(arxiv_id_or_url: str) -> str:
    """
    Return the arxiv id from a str which is either of the arxiv id itself or the url
    to the arxiv article.

    **Raises**
    - `ValueError`
        - If the input does not contain a valid arXiv ID.
    """
    new_id_pattern = r'(\d{4}\.\d{4,5}(?:v\d+)?)'
    old_id_pattern = r'([a-z-]+(?:\.[A-Z]{2})?/\d{7}(?:v\d+)?)'
    combined_pattern = f'({new_id_pattern}|{old_id_pattern})'
    
    # Check if input is a URL and extract the ID
    if 'arxiv.org' in arxiv_id_or_url:
        match = re.search(combined_pattern, arxiv_id_or_url)
        if match:
            return match.group(1)
        else:
            raise ValueError("Invalid arXiv URL provided.")
    
    # If it's not a URL, assume it's an ID and validate it
    elif re.match(combined_pattern, arxiv_id_or_url):
        return arxiv_id_or_url
    
    else:
        raise ValueError("Invalid input. Please provide a valid arXiv ID or URL.")


# %% ../../nbs/49_helper.arxiv.ipynb 8
def arxiv_search(
        arxiv_ids: Union[str, list[str]], # The ID of a single arXiv article or multiple arxiv articles
        client: Optional[Client] = None,  # an arxiv API Client. If `None`, create one on the spot.
        results: bool = True, # If `True` return a `Result` object. otherwise, return a `Search`` object`.
        ) -> Union[Result, Search]:
    if not client:
        client = Client()
    if not isinstance(arxiv_ids, list):
        arxiv_ids = [arxiv_ids]
    search = Search(id_list=arxiv_ids)
    if results:
        return client.results(search)
    return search

# %% ../../nbs/49_helper.arxiv.ipynb 14
def extract_metadata(
        results: Union[list[Result], Result],
        ) -> list[dict]: # Each dict corresponds to the metadata for each result.
    """
    Return the metadata from the arxiv search results
    """
    if not isinstance(results, list):
        results = [results]
    metadata_list = []
    for result in results:
        metadata = {
            "arxiv_id": result.get_short_id(),
            "authors": [author.name for author in result.authors],
            "title": result.title,
            "summary": result.summary,
            "primary_category": result.primary_category,
            "categories": result.categories,
            "published": result.published,
            "updated": result.updated,
            "doi": result.doi,
            "comment": result.comment,
            "journal_ref": result.journal_ref,
            "links": result.links
        }
        metadata_list.append(metadata)
    return metadata_list

# %% ../../nbs/49_helper.arxiv.ipynb 16
class ArxivMetadataEncoder(json.JSONEncoder):
    """
    `json` encoder to accomapny the `extract_metadta` function when using `json.dump`. 
    """
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.isoformat()
        elif isinstance(obj, arxiv.Result.Link):
            return obj.href
        return super().default(obj)

# %% ../../nbs/49_helper.arxiv.ipynb 19
def extract_last_names(
        authors: list[str]
        ) -> list[str]:
    last_names = []
    for author in authors:
        # Split the name into parts
        parts = author.split()
        # Handle special cases like "de Jong"
        if len(parts) > 2 and parts[-2].lower() in ['de', 'van', 'von', 'del', 'della', 'di', 'da', 'dos']:
            last_name = f"{parts[-2]} {parts[-1]}"
        else:
            last_name = parts[-1]
        
        # Remove any commas or periods
        last_name = re.sub(r'[,.]', '', last_name)
        last_names.append(last_name)
    return last_names


# %% ../../nbs/49_helper.arxiv.ipynb 22
def folder_name_for_source(
        result: Result,
        lowercase: bool = True
        ) -> str:
    family_names = extract_last_names([author.name for author in result.authors])
    if len(family_names) > 4:
        family_names_text = f'{family_names[0]}_et_al'
    else:
        family_names_text = '_'.join(family_names)
    if lowercase:
        output = f'{family_names_text.lower()}_{create_acronym(result.title)}'
    else:
        output = f'{family_names_text}_{create_acronym(result.title)}'
    return output


def create_acronym(title):
    # Words to exclude from acronym
    exclude_words = set(['a', 'an', 'the', 'on', 'and', 'of', 'to', 'over', 'in', 'for', 'with', 'by', 'at', 'from'])
    
    # Split the title into words
    words = re.findall(r'\b[\w-]+\b', title)
    
    acronym = ''
    for word in words:
        if word.lower() not in exclude_words:
            if len(word) == 1 and word.isupper():
                # Keep single uppercase letters (likely mathematical symbols) as is
                acronym += word
            elif word.lower() in ['i', 'ii', 'iii', 'iv', 'v', 'vi', 'vii', 'viii', 'ix', 'x']:
                # Handle Roman numerals
                acronym += word.lower()
            elif '-' in word:
                # Handle hyphenated words
                parts = word.split('-')
                acronym += ''.join(part[0].lower() if not (len(part) == 1 and part.isupper()) else part[0] for part in parts)
            else:
                # Take the first letter of other words, always lowercase
                acronym += word[0].lower()
    
    return acronym

# %% ../../nbs/49_helper.arxiv.ipynb 26
def file_name_for_pdf(
        result: Result
        ) -> str:
    family_names = extract_last_names([author.name for author in result.authors])
    if len(family_names) > 4:
        family_names_text = f'{family_names[0]} et al'
    else:
        family_names_text = ', '.join(family_names)
    output = f'{family_names_text} - {result.title}'
    return sanitize_filename(output)


# %% ../../nbs/49_helper.arxiv.ipynb 29
def analyze_arxiv_tarfile(
        filepath: PathLike # The path to the tar file.
        ) -> Literal["nested_archive", "direct_tex", "unknown_tar_structure", "plain_gz", "invalid_file"]:
    """
    Analyzes the contents of an arXiv download file, which can be either a tar.gz
    archive or a plain .gz file.

    This function attempts to determine the structure of the file downloaded from arXiv.
    It can identify several different types of content structures commonly found in arXiv
    downloads.

    Parameters:
    filepath (Union[str, Path]): The path to the file to be analyzed. Can be a string or a Path object.

    Returns:
    Literal["nested_archive", "direct_tex", "unknown_tar_structure", "plain_gz", "invalid_file"]: 
        - "nested_archive": If the tar.gz contains another compressed file
        - "direct_tex": If the tar.gz contains .tex files directly
        - "unknown_tar_structure": If the tar.gz structure doesn't match known patterns
        - "plain_gz": If the file is a plain .gz file (not a tar.gz)
        - "invalid_file": If the file is neither a valid tar.gz nor a valid .gz file

    Raises:
    No exceptions are raised; all errors are handled internally and returned as "invalid_file".
    Determine what kind of contents the 
    """
    try:
        with tarfile.open(filepath, "r:gz") as tar:
            members = tar.getmembers()
            
            if len(members) == 1 and members[0].name.endswith('.tar.gz'):
                return "nested_archive"
            
            tex_files = [m for m in members if m.name.endswith('.tex')]
            if tex_files:
                return "direct_tex"
            
            return "unknown_tar_structure"
    except tarfile.ReadError:
        # Check if it's a plain .gz file
        try:
            with gzip.open(filepath, 'rb') as gz_file:
                # Read a small portion to check if it's a valid gzip file
                gz_file.read(1024)
            return "plain_gz"
        except gzip.BadGzipFile:
            return "invalid_file"


def read_gz_file(filepath):
    with gzip.open(filepath, 'rt') as f:
        content = f.read()
    return content


def get_tex_filename_from_gz(filepath):
    with gzip.open(filepath, 'rt', encoding='utf-8') as f:
        content = f.read()
        
    # Look for a .tex filename in the content
    match = re.search(r'\b[\w-]+\.tex\b', content)
    if match:
        return match.group(0)
    else:
        return None


def extract_tex_from_gz(filepath):
    base_name = os.path.splitext(os.path.basename(filepath))[0]
    tex_filename = f"{base_name}.tex"
    
    with gzip.open(filepath, 'rb') as gz_file:
        content = gz_file.read()
    
    with open(tex_filename, 'wb') as tex_file:
        tex_file.write(content)
    
    return tex_filename

# %% ../../nbs/49_helper.arxiv.ipynb 31
def download_from_results(
        results: Result | list[Result],
        dir: PathLike, # The directory into which to download the files
        source: bool = True, # If `True`, download the source file. Otherweise, download a pdf file.
        # filename: Optional[str] = None, # The file name to save the file as. If `None`, then the filename is set to the arXiv id of the article.
        decompress_compressed_file: bool = True, # If `True`and if `source` is `True`, then decompress the source file after downloading it.
        file_or_folder_names: None | str | list[str] | Callable[Result, str] = folder_name_for_source, # If `None`, then the file/folder is named the arxiv id. If a `str` (in which case `results` must be a single `Result` or a `list[Result]` of length 1) or `list[str]` (whose length must equal that of `results`), then each file/folder is named by the specified corresponding `str`. If `Callable[Result, str]`, then each file/folder is named using the specified `Callable` 
        delete_compressed_file: bool = True, # If `True` and if `source` and `decompress_compressed_file` are `True`, then delete the compressed source file after downloading and then uncompressing it.
        download_metadata: bool = True, # If `True`, and if `source` is `True`, then create a file called `metadata.json` and put it into the newly created folder, unless a file called `metadata.json` already exists, in which case, a unique file name is created  
        verbose: bool = False,
        ) -> list[Path]: # Each `Path` is the folder in which the source files are newly downloaded or the path to the pdf file that is newly downloaded.
    """
    Download either the source files or pdfs of the arxiv article encoded in the results.

    - If `source = True` and `decompress_compressed_file = True`, then 
        - Download the source file/folder into a newly created folder (whose name is specified
          by `file_or_folder_names`) within `dir` and decompress the source (if applicable) in
          this newly created folder.
        - If `delete_compressed_file = True`, then delete the compressed file.
    - If `source = False`, then just download a pdf.

    For `file_or_folder_names`, the recommanded `Callable` arguments are `folder_name_for_source`
    for downloading source files and `file_name_for_pdf` for downloading pdf files.

    """
    if not isinstance(results, list):
        results = [results]
    if file_or_folder_names is not None and not isinstance(file_or_folder_names, (list,str)):
        file_or_folder_names = [file_or_folder_names(result) for result in results]
    elif file_or_folder_names is None:
        file_or_folder_names = [None for result in results]
    elif isinstance(file_or_folder_names, str):
        file_or_folder_names = [file_or_folder_names]
    downloaded_paths = []
    for result, file_or_folder_name in zip(results, file_or_folder_names):
        if not source:
            downloaded_paths.append(
                _download_pdf_with_name(result, dir, file_or_folder_name))
            continue
        downloaded_paths.append(
            _download_source(
                result, dir, file_or_folder_name, decompress_compressed_file,
                delete_compressed_file, download_metadata, verbose))
    return downloaded_paths



def _download_pdf_with_name(
        result: Result,
        dir: PathLike,
        file_or_folder_name: str|None # The name of the pdf file. If `None`, then the default name for the pdf file is used.
        ) -> Path:
    """
    Downloads the pdf for `result`.

    Helper function to `download_from_results`
    """
    if file_or_folder_name is not None:
        return Path(result.download_pdf(dir, filename=f'{file_or_folder_name}.pdf'))
    else:
        return Path(result.download_pdf(dir))
    

            

def _create_folder_for_source_download(
        result: Result,
        dir: PathLike,
        file_or_folder_name: str|None # The name of the folder in which the source should be downloaded. If `None`, then the arxiv id for `result` is used.
        ) -> Path: # The newly created folder 
    """
    Creates a new folder inside `dir` in which to download the source for `result`.

    Helper function to `download_from_results`
    """
    if file_or_folder_name is None:
        file_or_folder_name = result.entry_id
    new_folder = Path(dir) / file_or_folder_name
    if os.path.isdir(new_folder): #If folder exists
        # TODO: warn that folder exists
        if file_or_folder_name == result.entry_id:
            file_or_folder_name = f'{file_or_folder_name}_dupl'    
        else:
            file_or_folder_name = f'{file_or_folder_name}_{result.get_short_id()}'
        new_folder = Path(dir) / file_or_folder_name
    while os.path.isdir(new_folder): #If folder still exists
        file_or_folder_name = f'{file_or_folder_name}_dupl'
        new_folder = Path(dir) / file_or_folder_name
    os.mkdir(new_folder)
    return new_folder


def _download_source(
        result: Result,
        dir: PathLike,
        file_or_folder_name: str,
        decompress_compressed_file: bool,
        delete_compressed_file: bool,
        download_metadata: bool,
        verbose: bool
        ) -> Path:
    """
    Download source file into folder, decompress (as needed) the source, and download metadata

    Helper function to `download_from_results`
    """
    new_source_folder = _create_folder_for_source_download(result, dir, file_or_folder_name)
    source_file_path = result.download_source(new_source_folder)
    source_file_path = Path(new_source_folder) / source_file_path
    if verbose:
        print(source_file_path)
    compression_type = analyze_arxiv_tarfile(source_file_path)
    if decompress_compressed_file and file_is_compressed(source_file_path):
        if compression_type in ['plain_gz', 'direct_tex']:
            new_path = Path(f'{str(source_file_path)[:-7]}.gz')
            os.rename(source_file_path, new_path)
            source_file_path = new_path
        uncompressed = uncompress_file(source_file_path)
        if delete_compressed_file:
            os.remove(source_file_path)
        if len(uncompressed) == 1:
            if file_is_compressed(uncompressed[0]) and compression_type == 'nested_archive':
                uncompressed_again = uncompress_file(uncompressed[0])
                if delete_compressed_file:
                    os.remove(uncompressed[0])
            elif compression_type in ['plain_gz', 'direct_tex']:
                os.rename(uncompressed[0], f'{str(uncompressed[0])}.tex')
    if not download_metadata:
        return new_source_folder
    metadata_file_name = _unique_metadata_file_name(new_source_folder)
    metadata = extract_metadata(result)[0] 
    with open(new_source_folder / metadata_file_name, 'w') as json_file:
        json.dump(metadata, json_file, cls=ArxivMetadataEncoder, indent=4)
    return new_source_folder


def _unique_metadata_file_name(
        new_source_folder: PathLike # The folder in which to make the metadata file.
        ) -> str: #s tr: A unique file name for the metadata file within `new_source_folder`.

    """
    Identify a name to name the metadata file within `new_source_folder`; 
    the default name is `metadata.json` unless there is already a file with that name.
    
    If `metadata.json` exists, it appends a numeric suffix to create a unique file name 
    (e.g., `metadata_1.json`, `metadata_2.json`, etc.).
    """
    # Ensure the input is a Path object
    folder = Path(new_source_folder)
    
    # Default file name
    base_name = "metadata"
    extension = ".json"
    candidate = folder / f"{base_name}{extension}"
    
    # Check if the default file name exists
    counter = 1
    while candidate.exists():
        # TODO: warn that metadata.json exists
        # Generate a new candidate with a numeric suffix
        candidate = folder / f"{base_name}_{counter}{extension}"
        counter += 1
    
    return str(candidate.name)  # Return only the file name, not the full path




"""Helper functions for latex functionalities"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../../nbs/47_helper.latex.ipynb.

# %% auto 0
__all__ = ['REGEX_PATTERN_DETECTIONS', 'temp_dict', 'extract_latex_commands', 'extract_commands_from_nodes',
           'detect_incorrect_latex_commands', 'detect_unbalanced_environments',
           'math_mode_string_is_syntactically_valid', 'reduce_unnecessary_spaces', 'fix_autogen_formatting',
           'correct_latex_syntax_error']

# %% ../../../nbs/47_helper.latex.ipynb 3
import re
from typing import Callable, Union

from Levenshtein import distance
from pylatexenc.latexwalker import LatexCharsNode, LatexEnvironmentNode, LatexNode, LatexWalker, LatexMacroNode

from .. import sublist_generator
from ..regex import latex_indices, replace_string_by_indices
from .macros_and_commands import regex_pattern_detecting_command

# %% ../../../nbs/47_helper.latex.ipynb 9
def _does_not_end_with_script(s):
    """
    This is a helper function to `math_mode_string_is_syntactically_valid`.
    `s` is not supposed to have math mode delimits
    """
    s = s.strip(' $')
    return not s.endswith('_') and not s.endswith('^')

# %% ../../../nbs/47_helper.latex.ipynb 11
def _is_balanced_braces(s):
    """
    This is a helper function to `math_mode_string_is_syntactically_valid`.

    Note that curly braces (`{`, `}`) that are not preceded by backslashes
    '\\' are counted towards "balancing". 
    """
    stack = []
    escaped = False
    
    for _, char in enumerate(s):
        if char == '\\':
            escaped = True
        elif char == '{' and not escaped:
            stack.append(char)
        elif char == '}' and not escaped:
            if not stack:
                return False
            stack.pop()
        else:
            escaped = False
    
    return len(stack) == 0


# %% ../../../nbs/47_helper.latex.ipynb 13
def _first_curly_bracket(s) -> str|None:
    r"""
    Return whether a left curly bracket `{` or a right curly bracket `}`
    appears first in the string, ignoring escaped curly brackets `\{` or `\}`.
    """
    i = 0
    while i < len(s):
        if s[i] == '\\' and i + 1 < len(s):
            if s[i+1] in '{}':
                # Skip this escaped bracket
                i += 2
                continue
        elif s[i] == '{':
            return '{'
        elif s[i] == '}':
            return '}'
        i += 1
    return None


# %% ../../../nbs/47_helper.latex.ipynb 15
# def _last_curly_bracket(s) -> str|None:
def _last_curly_bracket(s):
    i = len(s) - 1
    while i >= 0:
        if s[i] in '{}':
            if i > 0 and s[i-1] == '\\':
                # This bracket is escaped
                i -= 2
                continue
            return s[i]
        i -= 1
    return None

# %% ../../../nbs/47_helper.latex.ipynb 17
def _detect_backslash_space_curly(
        text: str
        ) -> bool:
    """
    Return `True` if there is some backslash `\` followed
    by spaces and then followed by curly brackets `{`

    Note that the presence of such a combination of text
    will induce a syntax error in LaTeX math mode string.

    This is a helper function of `math_mode_string_is_syntactically_valid`
    """
    pattern = r'\\\s+[{}]'
    match = re.search(pattern, text)
    return bool(match)

# %% ../../../nbs/47_helper.latex.ipynb 19
def _is_left_right_balanced(
        latex_string: str
        ) -> bool:
    """
    Return `True` if occurrences of `\left` and `\right` are balanced. 

    This is a helper function of `math_mode_string_is_syntactically_valid`

    This function does not test whether occurrences of the
    appropriately corresponding braces are balanced. For instance,
    the function would return `True` on the input `\left . \right)`.
    Compare against `_is_semantically_left_right_balanced`, which
    is a similar function that tests whether left-right braces
    are "semantically" balanced.
    """
    # Remove all whitespace from the string
    latex_string = ''.join(latex_string.split())
    
    # Find all \left and \right commands
    left_commands = re.findall(r'\\left', latex_string)
    right_commands = re.findall(r'\\right', latex_string)
    
    # Check if the number of \left and \right commands are equal
    if len(left_commands) != len(right_commands):
        return False
    
    # Check if \left always comes before \right
    left_indices = [m.start() for m in re.finditer(r'\\left', latex_string)]
    right_indices = [m.start() for m in re.finditer(r'\\right', latex_string)]
    
    for left, right in zip(left_indices, right_indices):
        if left > right:
            return False
    
    return True

# %% ../../../nbs/47_helper.latex.ipynb 21
def _is_semantically_left_right_balanced(
        latex_string: str
        ) -> bool:
    """
    Return `True` if occurrences of `\left` and `\right` macros
    preceding various braces are balanced.

    This is a helper function of `math_mode_string_is_syntactically_clean`

    Compare against `_is_left_right_balanced`, which
    is a similar function that only tests whether left-right
    macros are "syntactically" balanced, without regard to
    the types of braces actually used.
    """
    # Remove all whitespace from the string
    latex_string = ''.join(latex_string.split())
    
    # Define a stack to keep track of opening delimiters
    stack = []
    
    # Define a dictionary to match opening and closing delimiters
    delimiters = {
        '(': ')', '[': ']', '{': '}', '<': '>',
        r'\left(': r'\right)', r'\left[': r'\right]', 
        r'\left{': r'\right}', r'\left<': r'\right>',
        r'\left\{': r'\right\}', r'\left|': r'\right|',
        r'\left\|': r'\right\|', r'\left.': r'\right.',
        r'\left\langle': r'\right\rangle'
    }
    
    # Regular expression to match \left and \right commands with their delimiters
    pattern = r'(\\left[\(\[\{\<\|\.\|]|\\right[\)\]\}\>\|\.\|]|\(|\)|\[|\]|\{|\}|\<|\>)'
    
    # Find all delimiters and \left/\right commands
    tokens = re.findall(pattern, latex_string)
    
    for token in tokens:
        if token.startswith(r'\left') or token in '([{<':
            stack.append(token)
        elif token.startswith(r'\right') or token in ')]}>':
            if not stack:
                return False
            last_open = stack.pop()
            if token != delimiters.get(last_open):
                return False
    
    # If the stack is empty, all delimiters are balanced
    return len(stack) == 0


# %% ../../../nbs/47_helper.latex.ipynb 23
def _has_invalid_left_right_bracket(
        latex_string: str
        ) -> bool:
    """
    Return `True` is there is at least one invalid use of
    a `\left` or `\right` command.

    This is a helper function of `math_mode_string_is_syntactically_valid`
    """
    # Remove all whitespace from the string
    latex_string = ''.join(latex_string.split())
    
    # Define valid brackets for \left and \right
    valid_brackets = [
        r'(', r')',
        r'[', r']',
        r'\(', r'\)', r'\[', r'\]',  # Parentheses and square brackets
        r'\{', r'\}',                # Curly braces (escaped)
        r'<', r'>',                  # Angle brackets
        r'\|',                       # Vertical bar
        r'|',                       # Vertical bar
        r'\\\|',                     # Double vertical bar (escaped)
        r'\.',                       # Dot
        r'.',                       # Dot
        r'\\lfloor', r'\\rfloor',    # Floor brackets
        r'\\lceil', r'\\rceil',      # Ceiling brackets
        r'\\langle', r'\\rangle'     # Angle brackets (commands)
    ]
    
    # Escape special regex characters and join with |
    valid_brackets_pattern = '|'.join(re.escape(b) for b in valid_brackets)
    
    # Pattern to match \left or \right followed by a valid bracket
    valid_pattern = rf'\\(left|right)({valid_brackets_pattern})'
    
    # Find all \left and \right commands
    commands = list(re.finditer(r'\\(left|right)', latex_string))
    
    for command in commands:
        # Check if the command is followed by a valid bracket
        if not re.match(valid_pattern, latex_string[command.start():]):
            # If not, return True and the invalid command
            invalid_part = latex_string[command.start():command.start()+6]  # Adjust slice as needed
            # return True, invalid_part
            return True
    
    # return False, None
    return False

# %% ../../../nbs/47_helper.latex.ipynb 25
def _has_double_script(
        latex_string: str
        ) -> bool:
    """
    Return `True` if there is at least one double superscript
    or double subscript in `latex_string` 

    This function fails to give correct outputs for more
    nuanced texts, such as `r"x^{2}_{3}^{4}"`; while in
    principle, the function should return `True` on this
    input, the actual return value is `False`.

    This is a helper function of `math_mode_string_is_syntactically_valid`
    """
    # Remove all whitespace from the string
    latex_string = ''.join(latex_string.split())
    
    # Function to match balanced braces
    def match_braces(s, start):
        count = 0
        for i, char in enumerate(s[start:], start):
            if char == '{':
                count += 1
            elif char == '}':
                count -= 1
                if count == 0:
                    return i
        return len(s) - 1

    # Find all subscripts and superscripts
    i = 0
    last_script = None
    while i < len(latex_string):
        if latex_string[i] in '^_' and (i == 0 or latex_string[i-1] != '\\'):
            current_script = latex_string[i]
            i += 1
            if i < len(latex_string):
                if latex_string[i] == '{':
                    end = match_braces(latex_string, i)
                    script_content = latex_string[i:end+1]
                    i = end + 1
                else:
                    script_content = latex_string[i]
                    i += 1
                
                if last_script and last_script[0] == current_script:
                    return True
                last_script = (current_script, script_content)
        else:
            if latex_string[i] not in '^_':
                last_script = None
            i += 1

    return False

# %% ../../../nbs/47_helper.latex.ipynb 27
def _has_double_script_literal(
        latex_string: str
        ) -> bool:
    """
    Return `True` if there is at least one double superscript
    or double subscript in `latex_string` by virtue of having
    `__`, `_^`, `^_, `^^`

    This is a helper function of `math_mode_string_is_syntactically_valid`
    """
    for bad_text in ['__', '_^', '^_', '^^']:
        if bad_text in latex_string:
            return True
    return False

# %% ../../../nbs/47_helper.latex.ipynb 29
def _has_unescaped_dollar(s):
    # Pattern explanation:
    # (?<!\\) is a negative lookbehind assertion that ensures the dollar sign is not preceded by a backslash
    # \$ matches a literal dollar sign
    pattern = r'(?<!\\)(\\\\)*\$'
    match = re.search(pattern, s)
    return bool(match)

# %% ../../../nbs/47_helper.latex.ipynb 32
def extract_latex_commands(latex_string):
    # Create a LatexWalker instance
    walker = LatexWalker(latex_string)
    
    # Get the nodes from the LaTeX string
    try:
        nodelist, _, _ = walker.get_latex_nodes()
    except Exception as e:
        print(f"Error parsing LaTeX: {e}")
        return []  # Return an empty list if there's a parsing error
    # Extract commands
    commands = []
    extract_commands_from_nodes(commands, nodelist)
    return commands


def extract_commands_from_nodes(
        commands: list[str],
        nodes: list[LatexNode]
        ):
    """
    This is a helper function to `extract_latex_commands`.
    """
    for node in nodes:
        # If the node is a character node, we skip it
        if isinstance(node, LatexCharsNode):
            continue
        elif isinstance(node, LatexMacroNode):
            commands.append(node.macroname)
            # Check for arguments of the macro node
            for arg in node.nodeargs:
                if arg and not isinstance(arg, LatexCharsNode):
                    if hasattr(arg, 'nodelist'):  # Ensure the argument is not None
                        extract_commands_from_nodes(commands, arg.nodelist)  # Extract from argument nodes
                    elif isinstance(arg, LatexMacroNode):
                        commands.append(arg.macroname)
        # elif isinstance(node, LatexEnvironmentNode):

        elif isinstance(node, LatexEnvironmentNode):
            commands.extend(_detect_begin_and_end_environments(node.latex_verbatim()))
        # If the node has a nodelist, extract commands from it
        if hasattr(node, 'nodelist'):
            extract_commands_from_nodes(commands, node.nodelist)

def _detect_begin_and_end_environments(
        latex_string: str
        ) -> list[str]:
    """
    Return a list of at most two items containing 'begin' if there is a \begin and containing 'end' if there is an \end

    This is a helper function to `extract_latex_commands`.
    """
    # Regular expressions to match \begin and \end with optional spaces
    begin_pattern = r'\\\s*begin'
    end_pattern = r'\\\s*end'
    
    # Initialize an empty result list
    result = []
    
    # Check for \begin
    if re.search(begin_pattern, latex_string):
        result.append('begin')
    
    # Check for \end
    if re.search(end_pattern, latex_string):
        result.append('end')
    
    return result

# %% ../../../nbs/47_helper.latex.ipynb 35
# Some arguments that can be used towards `regex_pattern_detecting_command`
# for some basic latex arguments.
# Note that the last argument doesn't actually matter, because
# we just want to be able to detect uses of comands, see
# `regex_pattern_detecting_commands``
REGEX_PATTERN_DETECTIONS = [
    ('frac', 2, None, None),
    ('binom', 2, None, None),
    ('sqrt', 1, '2', None),
    ('overset', 2, None, None),
    ('underset', 2, None, None),
    ('stackrel', 2, None, None),
    ('dfrac', 2, None, None),
    ('cfrac', 2, None, None),
    ('sideset', 3, None, None),
    ('xrightarrow', 1, None, None),
    ('xleftarrow', 1, None, None),
    ('overline', 1, None, None),
    ('bar', 1, None, None),
    ('arccos', 1, None, None),
    ('arcsin', 1, None, None),
    ('arctan', 1, None, None),
    ('arg', 1, None, None),
    ('atop', 2, None, None),
    ('begin', 1, None, None),
    ('boldsymbol', 1, None, None),
    ('breve', 1, None, None),
    ('check', 1, None, None),
    ('cline', 1, None, None),
    ('cos', 1, None, None),
    ('cosh', 1, None, None),
    ('cot', 1, None, None),
    ('csc', 1, None, None),
    ('dddot', 1, None, None),
    ('ddot', 1, None, None),
    ('dot', 1, None, None),
    ('end', 1, None, None),
    ('exp', 1, None, None),
    ('gcd', 2, None, None),
    ('grave', 1, None, None),
    ('hat', 1, None, None),
    # ('int', '1', None, None),
    ('lcm', 2, None, None),
    # ('left', 1, None, None),
    ('lg', 1, None, None),
    ('lim', 1, None, None),
    ('liminf', 1, None, None),
    ('limsup', 1, None, None),
    ('ln', 1, None, None),
    ('log', 1, None, None),
    ('longdiv', 2, None, None),
    ('lvert', 1, None, None),
    ('mapsto', 1, None, None),
    ('mathbb', 1, None, None),
    ('mathbf', 1, None, None),
    ('mathcal', 1, None, None),
    ('mathfrak', 1, None, None),
    ('mathop', 1, None, None),
    ('mathrm', 1, None, None),
    ('mathscr', 1, None, None),
    ('max', 1, None, None),
    ('min', 1, None, None),
    ('multicolumn', 3, 'center', None),
    ('multirow', 3, None, None),
    ('not', 1, None, None),
    ('oint', 1, None, None),
    ('overbrace', 1, None, None),
    ('overleftarrow', 1, None, None),
    ('overleftrightarrow', 1, None, None),
    ('overrightarrow', 1, None, None),
    # ('prod', 1, None, None),
    # ('right', 1, None, None),
    ('rvert', 1, None, None),
    ('sec', 1, None, None),
    ('section', 1, None, None),
    ('sin', 1, None, None),
    ('sinh', 1, None, None),
    ('stackrel', 2, None, None),
    ('subsection', 2, None, None),
    ('substack', 2, None, None),
    ('subsubsection', 2, None, None),
    # ('sum', 1, None, None),
    ('sup', 1, None, None),
    ('tag', 1, None, None),
    ('tan', 1, None, None),
    ('tanh', 1, None, None),
    ('text', 1, None, None),
    ('textbf', 1, None, None),
    ('textrm', 1, None, None),
    ('tilde', 1, None, None),
    ('underbrace', 1, None, None),
    ('underline', 1, None, None),
    ('underset', 2, None, None),
    ('varliminf', 1, None, None),
    ('varlimsup', 1, None, None),
    ('vec', 1, None, None),
    ('widehat', 1, None, None),
    ('widetilde', 1, None, None),
    ('xrightarrow', 1, None, None),
]
temp_dict = {}
for entry in REGEX_PATTERN_DETECTIONS:
    temp_dict[entry[0]] = entry
REGEX_PATTERN_DETECTIONS = temp_dict



# %% ../../../nbs/47_helper.latex.ipynb 36
def detect_incorrect_latex_commands(
        latex_string: str,
        ) -> bool:
    """
    Return `True` if there is at least one syntactically
    incorrect use of a latex command detected in `latex_string`.

    This is a helper function to `math_mode_string_is_syntactically_valid`.
    """
    commands_in_string = set(extract_latex_commands(latex_string))
    for command in commands_in_string:
        if command not in temp_dict:
            continue
        tuppy = temp_dict[command]
        pattern = regex_pattern_detecting_command(tuppy)
        # Look at each invocation of the command to see if 
        # each invocation is properly used.
        simp_pattern = rf"\\\s*{command}"
        simp_matches = re.finditer(simp_pattern, latex_string)
        # simp_matches = re.findall(simp_pattern, latex_string)
        for match in simp_matches:
            trailing_substring = latex_string[match.start():]
            alt_match = pattern.search(trailing_substring)
            if not alt_match or alt_match.span()[0] != 0:
                return True

        # if not matches and not simp_matches:
        #     continue
        # if len(matches) != len(simp_matches):
        #     return True
    return False

# %% ../../../nbs/47_helper.latex.ipynb 38
def detect_unbalanced_environments(
        latex_string: str) -> list[str]:
    # Define a regex pattern to match \begin{...} and \end{...}
    pattern = r'\\(begin|end)\{([^}]+)\}'
    
    # Stack to keep track of opened environments
    stack = []
    # List to store errors
    errors = []

    # Find all matches in the LaTeX string
    for match in re.finditer(pattern, latex_string):
        command, env_name = match.groups()
        
        if command == 'begin':
            # Push the environment name onto the stack
            stack.append(env_name)
        elif command == 'end':
            # Check if there is a matching begin for this end
            if stack and stack[-1] == env_name:
                stack.pop()  # Match found, pop from stack
            else:
                # Mismatch found, record the error
                errors.append(f"Mismatched \\end{{{env_name}}} at position {match.start()}")

    # If there are any unmatched begin commands left in the stack, report them
    while stack:
        unmatched_env = stack.pop()
        errors.append(f"Unmatched \\begin{{{unmatched_env}}}")

    return errors

# %% ../../../nbs/47_helper.latex.ipynb 40
def math_mode_string_is_syntactically_valid(
        text: str,
        ) -> bool:
    """
    Return `True` if `text` is determined to be syntactically valid
    as a latex str.

    There may be TeX syntax rules beyond the scope of this function.

    Some caveats:

    `text` is allowed to have dollar signs `$` and is also allowed to not have
    dollar signs. Even if `text` does not have dollar signs, this function
    may return `True`. Even if `text` has dollar signs, this function may return
    `False` if the entire string is not a singular math mode string or if the
    dollar signs are not used in a math-mode-valid way.
    """
    # 
    text = text.strip()
    math_mode_indices = latex_indices(text)
    if _has_unescaped_dollar(text):
        if len(math_mode_indices) != 1:
            return False
        if (math_mode_indices[0][0] != 0 or math_mode_indices[0][1] != len(text)):
            return False
    if not _does_not_end_with_script(text):
        return False
    if _detect_backslash_space_curly(text):
        return False
    if not _is_balanced_braces(text):
        return False
    if _has_invalid_left_right_bracket(text):
        return False
    if not _is_left_right_balanced(text):
        return False
    if _has_double_script(text):
        return False
    if _has_double_script_literal(text):
        return False
    if detect_incorrect_latex_commands(text):
        return False
    if bool(detect_unbalanced_environments(text)):
        return False
    return True



# %% ../../../nbs/47_helper.latex.ipynb 55
# def math_mode_string_is_syntactically_clean(
#         text: str,
#         ) -> bool:
#     """
#     Return `True` if `text` is syntactically "clean" as a LaTeX math mode str.
    
#     While the precise meaning of this may be subjective, here we will
#     consider `text` to be clean, assuming that it is syntactically valid, if

#     - It does not have double blackslashes
#     """
#     if r'\\' in text:
#         return False

# %% ../../../nbs/47_helper.latex.ipynb 58
def reduce_unnecessary_spaces(
        text: str,
        ) -> str:
    """
    Return a string modifying `text` by removing spaces which are
    unnecessary for the purposes of considering the string as a 
    LaTeX string.
    """
    pattern = r'([{_^\\()])\s+'
    text = re.sub(pattern, r'\1', text)
    pattern = r'\s+([}_^()])'
    text = re.sub(pattern, r'\1', text)
    return text
    # for char in ['{', '_', '^', '}', '\\']:
    #     text = re.sub(fr'\s*{chr}\s*', chr, text)

# %% ../../../nbs/47_helper.latex.ipynb 61
def fix_autogen_formatting(
        text: str
        ) -> str:
    """Fix some latex formatting issues in an autogenerated text
    """
    text = text.replace(r'\ ', '\\')
    text = text.replace(r'{ ', r'{')
    text = text.replace(r' }', r'}')
    text, _ = re.subn(r'\$\s*([^\$]+?)\s*\$', r'$\1$', text)
    # TODO: do $ <latex_string> $ into $<latex_stinrg>$
    # TODO: if the replacement of r'\ ' by '\\' happesn to
    # make `\` stick to the previous chunk of things
    # (e.g. r'd\in\mathbb{Z}_{\geq 0}`, then give it some
    # space, e.g. r'd \in \mathbb{Z}_{\geq 0}'.
    text = reduce_unnecessary_spaces(text)
    text = _insert_newline_or_spaces_around_latex(text)
    return text


def _insert_newline_or_spaces_around_latex(
        text:  str
        ) -> str:
    """
    Insert spaces or newlines around latex math mode strings inside `text`
    if necessary.
    """
    math_mode_indices = latex_indices(text)
    replacements = []
    for start, end in math_mode_indices:
        math_mode = text[start:end]
        spaces_potentially_added = math_mode
        if not math_mode.startswith('$$'): #starts with exactly one $
            if start != 0 and text[start-1] != ' ':
                spaces_potentially_added = f' {spaces_potentially_added}'
            if end != len(text) and text[end] != ' ':
                spaces_potentially_added = f'{spaces_potentially_added} '
            replacements.append(spaces_potentially_added)
            continue
        if start != 0 and text[start-1] != '\n':
            front_newline_count = 2
        elif start > 1 and text[start-2] != '\n':
            front_newline_count = 1
        else:
            front_newline_count = 0
        spaces_potentially_added = front_newline_count * '\n' + spaces_potentially_added

        if end != len(text) and text[end] != '\n':
            back_newline_count = 2
        elif end < len(text) - 1 and text[end-1] != '\n':
            back_newline_count = 1
        else:
            back_newline_count = 0
        spaces_potentially_added = spaces_potentially_added + '\n'*back_newline_count
        replacements.append(spaces_potentially_added)
    text = replace_string_by_indices(text, math_mode_indices, replacements)
    text = text.replace('$  $', '$ $')
    text = text.replace('$$\n\n\n\n$$', '$$\n\n$$')
    text = text.replace('$$\n\n\n$$', '$$\n\n$$')
    return text

# %% ../../../nbs/47_helper.latex.ipynb 68
def _tokenize_latex_math(
        latex_string: str
        ) -> list[str]:
    """
    Tokenize `latex_string` by the following principles:

    1. A latex command/macro invoked (but not the inputs) is a token.
    2. the special characters ^ { } _ are tokens.
    3. groups of consecutive whitespaces are tokens.
    4. afterwards, all "words" (one or more consecutive non-whitespace non-special characters) are tokens.
    """
    # Define the regex pattern for tokenization
    pattern = r"""
        (\\[a-zA-Z]+)        # Match LaTeX commands (e.g., \alpha, \sum)
        | ([^\\\s^{}_]+)     # Match words (consecutive non-whitespace, non-special characters)
        | ([^\\\s])          # Match special characters (including ^, {, }, _, etc.)
        | (\s+)              # Match groups of consecutive whitespace
    """
    # Use re.findall to find all matches based on the pattern
    tokens = re.findall(pattern, latex_string, re.VERBOSE)
    # Extract the matched groups, filtering out empty strings
    token_list = [token for group in tokens for token in group if token]
    return token_list


# %% ../../../nbs/47_helper.latex.ipynb 70
def _list_of_candidates_from_math_mode_strings(
        main_content: str, # A text of LaTeX code. In practice, this should be the `main content` of an information note, cf. `summarize_notation`.`
        syntax_validation: Callable[str, bool] = math_mode_string_is_syntactically_valid # A test to tell whether a math mode string is syntactically  valid.
        ) -> set[str]:
    """
    Return a substrings from latex math mode strings in `main_content`
    that are syntactically valid .

    None of the elements in the output have delimiters (`$`, `$$`)
    """
    syntactically_valid_substrings = [] 
    math_mode_indices = latex_indices(main_content)
    for start, end in math_mode_indices:
        latex_str = main_content[start:end]
        latex_str = latex_str.strip('$')
        tokenization = _tokenize_latex_math(latex_str)
        for sublist in sublist_generator(tokenization):
            substring = ''.join(sublist)
            if syntax_validation(substring):
                syntactically_valid_substrings.append(substring.strip())
    return set(syntactically_valid_substrings)

# %% ../../../nbs/47_helper.latex.ipynb 73
def _find_closest_match(
        math_mode_text: str,
        replacement_candidates: list[str]
        ) -> Union[str, None]:
    """This is a helper function to `correct_latex_syntax_error`."""
    if not replacement_candidates:
        return None
    # Calculate Levenshtein distance for each candidate
    distances = [(candidate, distance(math_mode_text, candidate)) for candidate in replacement_candidates]
    # Find the candidate with the minimum distance
    closest_match = min(distances, key=lambda x: x[1])
    return closest_match[0]

# %% ../../../nbs/47_helper.latex.ipynb 75
def correct_latex_syntax_error(
        summary: str, # The autogenerated summary
        replacement_candidates: list[str], # A list of candidates to replace. This is expected to be an output of `_list_of_candidates_from_math_mode_strings`
        # min_length_to_replace_math_mode_string: int = 5, # The minimum length that a math mode string needs to be (exclusing delimiting dollar signs `$`, `$$`) in summary in order to be considered for replacement.
        syntax_validation: Callable[str, bool] = math_mode_string_is_syntactically_valid # A test to tell whether a math mode string is syntactically  valid.
        ) -> str:
    """
    Attempt to replace within `summary` a modified version in which
    the syntactically incorrect latex math mode strings are replaced
    with the most closely resembling element of `replacement_candidates`. 
     
    with a modified version in which the
    latex math mode strings within `summary` that are syntactically
    incorrect 

    TODO: consider the possibility that not all math mode str delimiters
    are formatted correctly.
    """
    math_mode_indices = latex_indices(summary)
    replacements = []
    for start, end in math_mode_indices:
        math_mode_text = summary[start:end]
        if syntax_validation(math_mode_text) or not replacement_candidates:
            replacements.append(math_mode_text)
            continue
        delimiter = '$$' if math_mode_text.startswith('$$') else '$'
        replacement = _find_closest_match(math_mode_text, replacement_candidates)
        replacement = f'{delimiter}{replacement}{delimiter}'
        replacements.append(replacement)
    return replace_string_by_indices(summary, math_mode_indices, replacements)


